# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import OrderedDict
from copy import deepcopy
from typing import Callable, Dict, List, Tuple

import numpy as np
import torch

import nncf
from nncf.common.graph.layer_attributes import WeightedLayerAttributes
from nncf.common.quantization.initialization.range import RangeInitCollectorParams
from nncf.common.quantization.initialization.range import RangeInitConfig
from nncf.common.quantization.initialization.range import RangeInitParams
from nncf.common.quantization.quantizer_setup import QuantizationPointBase
from nncf.common.quantization.quantizer_setup import QuantizerSetupBase
from nncf.common.quantization.structs import NonWeightQuantizerId
from nncf.common.quantization.structs import QuantizationScheme
from nncf.common.quantization.structs import QuantizerGroup
from nncf.common.quantization.structs import QuantizerId
from nncf.common.quantization.structs import WeightQuantizerId
from nncf.common.scopes import should_consider_scope
from nncf.common.tensor_statistics.collectors import ReductionAxes
from nncf.common.tensor_statistics.collectors import TensorStatisticCollectorBase
from nncf.config.schemata.algo.quantization import RANGE_INIT_TYPES_VS_DESCRIPTIONS
from nncf.experimental.common.tensor_statistics.collectors import AggregationAxes
from nncf.torch.graph.graph import PTNNCFGraph
from nncf.torch.initialization import DataLoaderBaseRunner
from nncf.torch.nncf_network import NNCFNetwork
from nncf.torch.quantization.layers import BaseQuantizer
from nncf.torch.quantization.layers import SymmetricQuantizer
from nncf.torch.quantization.layers import get_scale_shape
from nncf.torch.quantization.translator import PTTargetPointTranslator
from nncf.torch.tensor_statistics.algo import TensorStatisticObservationPoint
from nncf.torch.tensor_statistics.algo import create_register_input_hook
from nncf.torch.tensor_statistics.collectors import get_mean_percentile_statistic_collector
from nncf.torch.tensor_statistics.collectors import get_median_mad_statistic_collector
from nncf.torch.tensor_statistics.collectors import get_min_max_statistic_collector
from nncf.torch.tensor_statistics.collectors import get_mixed_min_max_statistic_collector
from nncf.torch.tensor_statistics.collectors import get_percentile_tensor_collector
from nncf.torch.tensor_statistics.statistics import pt_convert_stat_to_min_max_tensor_stat


class PTRangeInitParams(RangeInitParams):
    def get_max_num_init_steps(self) -> int:
        steps = []
        if self.global_init_config is not None:
            steps.append(self.global_init_config.num_init_samples)
        for pl_config in self.per_layer_range_init_configs:
            steps.append(pl_config.num_init_samples)
        batch_size = self.init_range_data_loader.batch_size
        return int(np.ceil(max(steps) / batch_size))

    def get_init_config_for_quantization_point(self, qp: QuantizationPointBase) -> RangeInitConfig:
        if qp.is_weight_quantization_point():
            qid = WeightQuantizerId(qp.insertion_point.target_node_name)
            group = QuantizerGroup.WEIGHTS
        else:
            qid = NonWeightQuantizerId(qp.insertion_point.target_node_name, qp.insertion_point.input_port_id)
            group = QuantizerGroup.ACTIVATIONS
        return self.get_init_config_for_scope_and_group(qid, group)

    def get_init_config_for_scope_and_group(self, qid: QuantizerId, group: QuantizerGroup) -> RangeInitConfig:
        matches: List[RangeInitConfig] = []
        for pl_config in self.per_layer_range_init_configs:
            should_be_considered = should_consider_scope(qid, pl_config.ignored_scopes, pl_config.target_scopes)
            if should_be_considered and (group == pl_config.target_group or pl_config.target_group is None):
                matches.append(
                    RangeInitConfig(
                        pl_config.init_type, pl_config.num_init_samples, pl_config.init_type_specific_params
                    )
                )
        if len(matches) > 1:
            raise ValueError(
                "Location {} matches more than one per-layer initialization parameter definition!".format(str(qid))
            )
        if len(matches) == 1:
            return matches[0]
        if not matches and self.global_init_config is not None:
            return deepcopy(self.global_init_config)

        raise ValueError(
            "Location {} does not match any per-layer initialization parameter definition!".format(str(qid))
        )


class PTRangeInitCollectorParams(RangeInitCollectorParams):
    def __init__(
        self, is_weights: bool, scheme: QuantizationScheme, per_channel: bool, input_shape: tuple, channel_idx: int
    ):
        """
        :param is_weights: Boolean that defines tensor type. True for Weights, False for Activations.
        :param scheme: Quantization scheme: symmetric or asymmetric.
        :param input_shape: Shape of the input tensor.
        :param channel_idx: Channel dimension.
        """
        super().__init__(is_weights, scheme, per_channel)
        self._input_shape = input_shape
        self._channel_idx = channel_idx

    def get_reduction_aggregation_axes(self, is_per_sample: bool) -> Tuple[ReductionAxes, AggregationAxes]:
        if self.is_per_channel:
            return super().get_reduction_aggregation_axes(self._input_shape, (self._channel_idx,), is_per_sample)
        return super().get_reduction_aggregation_axes(self._input_shape, (), is_per_sample)


class StatCollectorGenerator:
    @staticmethod
    def generate_collectors_for_range_init_statistics_collection(
        target_model_graph: PTNNCFGraph, quantizer_setup: QuantizerSetupBase, range_init_params: PTRangeInitParams
    ) -> Dict[TensorStatisticObservationPoint, Dict[ReductionAxes, TensorStatisticCollectorBase]]:
        retval = {}
        for qp in quantizer_setup.quantization_points.values():
            init_config = range_init_params.get_init_config_for_quantization_point(qp)
            is_weights = qp.is_weight_quantization_point()
            num_batches = int(
                np.ceil(init_config.num_init_samples / range_init_params.init_range_data_loader.batch_size)
            )
            if is_weights:
                # No need to store extra statistics in memory since weights won't change during range init
                num_batches = 1

            tp = PTTargetPointTranslator.translate(qp.insertion_point)
            scale_shapes_vs_params = StatCollectorGenerator.get_all_scale_shapes_with_params(qp, target_model_graph)

            obs_p = TensorStatisticObservationPoint(tp, reduction_shapes=set(scale_shapes_vs_params.keys()))

            retval[obs_p] = {}
            for scale_shape in obs_p.reduction_shapes:
                collector_params = scale_shapes_vs_params[scale_shape]
                collector = StatCollectorGenerator.generate_stat_collector_for_range_init_config(
                    init_config, scale_shape, collector_params, num_samples_to_collect_override=num_batches
                )
                retval[obs_p][scale_shape] = collector

        return retval

    @staticmethod
    def generate_stat_collector_for_range_init_config(
        init_config: RangeInitConfig,
        scale_shape: ReductionAxes = None,
        collector_params: PTRangeInitCollectorParams = None,
        num_samples_to_collect_override: int = None,
    ) -> TensorStatisticCollectorBase:
        num_samples = init_config.num_init_samples
        if num_samples_to_collect_override is not None:
            num_samples = num_samples_to_collect_override
        if init_config.init_type not in RANGE_INIT_TYPES_VS_DESCRIPTIONS:
            raise nncf.InternalError("Unknown range init type: {}".format(init_config.init_type))

        use_per_sample_stats = collector_params.use_per_sample_stats(init_config.init_type == "mixed_min_max")
        reduction_axes, aggregation_axes = collector_params.get_reduction_aggregation_axes(use_per_sample_stats)
        if init_config.init_type == "min_max":
            return get_min_max_statistic_collector(
                use_abs_max=collector_params.use_abs_max,
                reduction_axes=reduction_axes,
                aggregation_axes=aggregation_axes,
                scale_shape=scale_shape,
                num_samples=num_samples,
            )
        if init_config.init_type == "mixed_min_max":
            return get_mixed_min_max_statistic_collector(
                use_abs_max=collector_params.use_abs_max,
                reduction_axes=reduction_axes,
                aggregation_axes=aggregation_axes,
                scale_shape=scale_shape,
                use_means_of_mins=collector_params.use_means_of_mins,
                use_means_of_maxs=collector_params.use_means_of_maxs,
                num_samples=num_samples,
            )
        if init_config.init_type == "mean_min_max":
            return get_mixed_min_max_statistic_collector(
                use_abs_max=collector_params.use_abs_max,
                reduction_axes=reduction_axes,
                aggregation_axes=aggregation_axes,
                scale_shape=scale_shape,
                use_means_of_mins=True,
                use_means_of_maxs=True,
                num_samples=num_samples,
            )
        if init_config.init_type == "threesigma":
            return get_median_mad_statistic_collector(
                reduction_axes=reduction_axes,
                aggregation_axes=aggregation_axes,
                scale_shape=scale_shape,
                num_samples=num_samples,
            )
        if init_config.init_type == "percentile":
            min_percentile = init_config.init_type_specific_params.get("min_percentile", 0.1)
            max_percentile = init_config.init_type_specific_params.get("max_percentile", 99.9)
            return get_percentile_tensor_collector(
                percentiles_to_collect=(min_percentile, max_percentile),
                reduction_axes=reduction_axes,
                aggregation_axes=aggregation_axes,
                scale_shape=scale_shape,
                num_samples=num_samples,
            )

        if init_config.init_type == "mean_percentile":
            min_percentile = init_config.init_type_specific_params.get("min_percentile", 0.1)
            max_percentile = init_config.init_type_specific_params.get("max_percentile", 99.9)
            return get_mean_percentile_statistic_collector(
                percentiles_to_collect=(min_percentile, max_percentile),
                reduction_axes=reduction_axes,
                aggregation_axes=aggregation_axes,
                scale_shape=scale_shape,
                num_samples=num_samples,
            )
        raise ValueError("Range init type not handled!")

    @classmethod
    def get_all_scale_shapes_with_params(
        cls, qp: QuantizationPointBase, target_nncf_graph: PTNNCFGraph
    ) -> Dict[ReductionAxes, PTRangeInitCollectorParams]:
        qconfigs = qp.get_all_configs_list()
        if qp.is_weight_quantization_point():
            module_node = target_nncf_graph.get_node_by_name(qp.insertion_point.target_node_name)
            layer_attributes = module_node.layer_attributes
            assert isinstance(layer_attributes, WeightedLayerAttributes)
            input_shape = layer_attributes.get_weight_shape()
            channel_idx = layer_attributes.get_target_dim_for_compression()
        else:
            input_shape = target_nncf_graph.get_input_shape_for_insertion_point(qp.insertion_point)
            channel_idx = 1  # channel dim for activations

        retval = {}
        for qconfig in qconfigs:
            is_weights = qp.is_weight_quantization_point()
            scale_shape = tuple(
                get_scale_shape(
                    input_shape, is_weights=is_weights, per_channel=qconfig.per_channel, channel_idx=channel_idx
                )
            )

            if scale_shape not in retval:
                retval[scale_shape] = PTRangeInitCollectorParams(
                    is_weights, qconfig.mode, qconfig.per_channel, input_shape, channel_idx
                )
        return retval


class DataLoaderRangeInitializeRunner(DataLoaderBaseRunner):
    def __init__(
        self,
        model: NNCFNetwork,
        modules_to_init_vs_init_configs: Dict[str, Tuple[BaseQuantizer, RangeInitConfig, bool, Tuple[int]]],
        init_device: str,
        batch_size: int = None,
    ):
        super().__init__(model, init_device)
        self.modules_to_init = modules_to_init_vs_init_configs
        self.progressbar_description = "Range parameters initialization"

        self.collectors_and_modules_to_init: Dict[str, Tuple[TensorStatisticCollectorBase, BaseQuantizer]] = (
            OrderedDict()
        )
        self.hook_handles = []
        self.batch_size = batch_size

    def _get_fwd_hook(
        self, collector: TensorStatisticCollectorBase
    ) -> Callable[["torch.Module", torch.Tensor, torch.Tensor], torch.Tensor]:
        hook = create_register_input_hook(collector=collector)

        def fwd_hook(module, input_, output):
            hook(input_[0])

        return fwd_hook

    def _prepare_initialization(self):
        for name, data in self.modules_to_init.items():
            quantizer_module, init_config, is_weights, input_shape = data
            num_samples_override = None
            if self.batch_size is not None:
                num_batches = np.ceil(init_config.num_init_samples / self.batch_size)
                num_samples_override = num_batches

            if isinstance(quantizer_module, SymmetricQuantizer):
                mode = QuantizationScheme.SYMMETRIC
            else:
                mode = QuantizationScheme.ASYMMETRIC

            shape = quantizer_module.scale_shape
            if shape == (1,):  # Per-tensor
                channel_idx = None
            elif len(shape) > 1 and all(item == 1 for item in shape):
                channel_idx = 0  # (1, 1, 1, 1) - doest not matter which dim is channel_idx
            else:
                if not is_weights:
                    channel_idx = 1  # channel dim for activations
                else:
                    channel_idx = [i for i, val in enumerate(shape) if val != 1][0]

            collector_params = PTRangeInitCollectorParams(
                is_weights, mode, quantizer_module.per_channel, input_shape, channel_idx
            )

            collector = StatCollectorGenerator.generate_stat_collector_for_range_init_config(
                init_config, tuple(quantizer_module.scale_shape), collector_params, num_samples_override
            )

            self.collectors_and_modules_to_init[name] = collector, quantizer_module

            self.hook_handles.append(quantizer_module.register_forward_hook(self._get_fwd_hook(collector)))

    def _apply_initializers(self):
        for handle in self.hook_handles:
            handle.remove()
        for scope_str, collector_and_module in self.collectors_and_modules_to_init.items():
            collector, quantizer_module = collector_and_module
            target_stat = collector.get_statistics()
            minmax_stats = pt_convert_stat_to_min_max_tensor_stat(target_stat)
            quantizer_module.apply_minmax_init(
                minmax_stats.min_values, minmax_stats.max_values, log_module_name=scope_str
            )



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from math import inf
from typing import Any, Dict, List, Optional, Tuple, TypeVar, Union

import nncf
from nncf import Dataset
from nncf.common.factory import EngineFactory
from nncf.common.factory import ModelTransformerFactory
from nncf.common.graph.graph import NNCFGraph
from nncf.common.graph.model_transformer import ModelTransformer
from nncf.common.graph.transformations.commands import TargetPoint
from nncf.common.graph.transformations.commands import TargetType
from nncf.common.graph.transformations.layout import TransformationLayout
from nncf.common.logging import nncf_logger
from nncf.common.logging.track_progress import track
from nncf.common.tensor_statistics.statistic_point import StatisticPoint
from nncf.common.tensor_statistics.statistic_point import StatisticPointsContainer
from nncf.common.utils.backend import BackendType
from nncf.common.utils.backend import get_backend
from nncf.experimental.common.tensor_statistics.statistical_functions import mean_per_channel
from nncf.experimental.tensor import Tensor
from nncf.experimental.tensor import functions as fns
from nncf.quantization.algorithms.algorithm import Algorithm

TModel = TypeVar("TModel")
TTensor = TypeVar("TTensor")

FAST_BIAS_CORRECTION_THRESHOLD = 2


class FastBiasCorrection(Algorithm):
    """
    Post-training FastBiasCorrection algorithm implementation.

    The main purpose of this algorithm to reduce quantization error
    via correction the bias of the Convolutions, FullyConnected, etc. layers.
    The algorithm pipeline is very simple:
        - we collects floating-point statistics from the corresponding model for the layers with bias;
        - then we gets the quantized model and try to reduce it's error by correction of the bias;
        - the shift calculates using the sub-graph that consists of the correction layer and
        weight quantizer-dequantizer pair or fake quantize node;
        - the floating-point statistics uses as input for
        the sub-graph and further quantization output calculation;
        - in the end we corrects the original bias by the difference (shift)
        between floating-point and quantized outputs.
    """

    def __init__(
        self,
        subset_size: int = 100,
        threshold: float = FAST_BIAS_CORRECTION_THRESHOLD,
        apply_for_all_nodes: bool = False,
        inplace_statistics: bool = True,
        backend_params: Optional[Dict[str, Any]] = None,
    ):
        """
        :param subset_size: Size of a subset for the statistics collection,
            defaults to 100.
        :param threshold: The magnitude threshold that regulates the application of the
            shift. Magnitude calculates as the maximum of the absolute ratio of the
            shift to the original bias value. If the calculated value is less than the
            threshold, the shift will apply to the bias, defaults to 2.
        :param apply_for_all_nodes: If True, then the bias correction be applied to all
            quantized nodes, if the node has no bias then a bias node will be inserted,
            and if False, then the bias correction will only be applied to quantized
            nodes that have a bias.
        :param inplace_statistics: Defines wheather to calculate quantizers statistics
            by backend graph operations or by default Python implementation, defaults
            to True.
        :param backend_params: Backend specific parameters.
        """
        super().__init__()
        self.subset_size = subset_size
        self.threshold = threshold
        self.apply_for_all_nodes = apply_for_all_nodes
        self.inplace_statistics = inplace_statistics
        self.backend_params = backend_params
        self._backend_entity = None
        self._algorithm_key = f"FBC_{hash(self)}"

        if self.apply_for_all_nodes:
            raise nncf.InternalError("FastBiasCorrection algorithm does not support apply_for_all_nodes=True yet")

    @property
    def available_backends(self) -> List[BackendType]:
        return [BackendType.ONNX, BackendType.OPENVINO, BackendType.TORCH]

    def _set_backend_entity(self, model: TModel) -> None:
        """
        Creates a helper class with a backed-specific logic of the algorithm.

        :param model: Backend-specific input model.
        """
        model_backend = get_backend(model)
        if model_backend == BackendType.ONNX:
            from nncf.quantization.algorithms.fast_bias_correction.onnx_backend import ONNXFastBiasCorrectionAlgoBackend

            self._backend_entity = ONNXFastBiasCorrectionAlgoBackend()
        elif model_backend == BackendType.OPENVINO:
            from nncf.quantization.algorithms.fast_bias_correction.openvino_backend import (
                OVFastBiasCorrectionAlgoBackend,
            )

            self._backend_entity = OVFastBiasCorrectionAlgoBackend()
        elif model_backend == BackendType.TORCH:
            from nncf.quantization.algorithms.fast_bias_correction.torch_backend import PTFastBiasCorrectionAlgoBackend

            self._backend_entity = PTFastBiasCorrectionAlgoBackend()
        else:
            raise nncf.UnsupportedBackendError(
                "Cannot return backend-specific entity because {} is not supported!".format(model_backend.value)
            )

    def apply(
        self,
        model: TModel,
        graph: NNCFGraph,
        statistic_points: Optional[StatisticPointsContainer] = None,
        dataset: Optional[Dataset] = None,
    ) -> TModel:
        self._set_backend_entity(model)

        model_transformer = ModelTransformerFactory.create(model)

        node_and_bias_value = [
            (node, self._backend_entity.get_bias_value(node, graph, model))
            for node in graph.get_all_nodes()
            if self._backend_entity.is_node_with_bias(node, graph)
        ]

        # Fill `node_and_new_bias_value` list. It is a correspondence between nodes
        # for which we should update bias and new bias values.
        node_and_new_bias_value = []

        for node, bias_value in track(node_and_bias_value, description="Applying Fast Bias correction"):
            node_name = node.node_name

            if not self._backend_entity.is_quantized_weights(node, graph):
                nncf_logger.debug(f"Skipping node {node_name} because weights were not quantized")
                continue

            in_node_name, out_node_name = self._backend_entity.get_node_names_for_input_output_statistics(node, graph)

            input_fp, input_shape = self._get_fp_inputs(statistic_points, in_node_name)
            output_fp = self._get_fp_outputs(statistic_points, out_node_name)

            extracted_model = self._extract_submodel(model_transformer, in_node_name, out_node_name)
            if extracted_model is None:
                nncf_logger.debug(f"Skipping node {node_name} because cant extract submodel")
                continue

            sub_input_name, sub_output_name = self._backend_entity.get_sub_input_output_names(extracted_model)

            channel_axis = node.metatype.output_channel_axis
            if bias_value.ndim > 1:
                # Make index positive
                channel_axis = range(bias_value.ndim)[channel_axis]
            input_blob = self._backend_entity.create_input_data(input_shape, input_fp, sub_input_name, channel_axis)
            bias_shift = self._get_bias_shift(
                model=extracted_model,
                input_blob=input_blob,
                channel_axis=channel_axis,
                output_fp=output_fp,
                output_name=sub_output_name,
            )

            bias_shift = self._reshape_bias_shift(bias_shift, bias_value, channel_axis)
            updated_bias = bias_value + bias_shift
            magnitude = self._get_bias_shift_magnitude(bias_value, updated_bias)

            if magnitude < self.threshold:
                nncf_logger.debug(f"{node_name} bias would be changed")
                node_and_new_bias_value.append((node, updated_bias))
            else:
                nncf_logger.debug(f"{node_name} bias skipped by threshold. Magnitude: {magnitude}")

        # Create commands of bias correction and apply them to the model.
        transformation_layout = TransformationLayout()
        for node, bias_value in node_and_new_bias_value:
            transformation_layout.register(self._backend_entity.create_bias_correction_command(node, bias_value, graph))
        transformed_model = model_transformer.transform(transformation_layout)

        return transformed_model

    @staticmethod
    def _get_bias_shift_magnitude(current_bias_value: Tensor, updated_bias_value: Tensor) -> float:
        """
        Calculates bias shift magnitude based on the current and updated values.

        :param current_bias_value: The original bias value.
        :param updated_bias_value: The updated bias value.
        :return: Magnitude between original and updated bias values.
        """
        bias_shift_magnitude = inf
        if fns.count_nonzero(current_bias_value == 0) == 0:
            bias_shift_magnitude = fns.max(fns.abs((updated_bias_value - current_bias_value) / current_bias_value))
        return bias_shift_magnitude

    @staticmethod
    def _reshape_bias_shift(bias_shift: Tensor, bias_value: Tensor, channel_axis: int) -> Tensor:
        """
        Reshape bias_shift tensor in case of dimensions of bias_value is more then 1.

        :param bias_shift: Bias shift tensor.
        :param bias_value: Bias value tensor.
        :param channel_axis: Axis to update bias.

        :return TTensor: Updated bias_shift.
        """
        if bias_value.ndim > 1:
            new_shape = [1] * bias_value.ndim
            new_shape[channel_axis] = bias_shift.shape[0]
            bias_shift = bias_shift.reshape(new_shape)
        return bias_shift

    def _get_fp_inputs(self, statistic_points: StatisticPointsContainer, node_name: str) -> Tuple[List, List]:
        """
        Makes out per-layer needed data from the floating-point collected statistics.

        :param statistic_points: Filled StatisticPointsContainer.
        :param node_name: Name of the current layer.
        :return: Collected mean tensor data and shape for the further bias calculation.
        """

        def input_filter_func(point):
            return self._algorithm_key in point.algorithm_to_tensor_collectors and point.target_point.type in [
                TargetType.PRE_LAYER_OPERATION,
                TargetType.OPERATOR_PRE_HOOK,
            ]

        input_fp = []
        input_shape = []
        for tensor_collector in statistic_points.get_algo_statistics_for_node(
            node_name, input_filter_func, self._algorithm_key
        ):
            statistics = tensor_collector.get_statistics()
            input_fp.extend(Tensor(statistics.mean_values))
            input_shape.extend(statistics.shape)
        return input_fp, input_shape

    def _get_fp_outputs(self, statistic_points: StatisticPointsContainer, node_name: str) -> List[TTensor]:
        """
        Makes out per-layer needed data from the floating-point collected statistics.

        :param statistic_points: Filled StatisticPointsContainer.
        :param node_name: Name of the current layer.
        :return: Collected mean tensor data for the further bias calculation.
        """

        def output_filter_func(point):
            return self._algorithm_key in point.algorithm_to_tensor_collectors and point.target_point.type in [
                TargetType.POST_LAYER_OPERATION,
                TargetType.OPERATOR_POST_HOOK,
            ]

        output_fp = []
        for tensor_collector in statistic_points.get_algo_statistics_for_node(
            node_name, output_filter_func, self._algorithm_key
        ):
            output_fp.extend(Tensor(tensor_collector.get_statistics().mean_values))
        return output_fp

    def _extract_submodel(self, model_transformer: ModelTransformer, in_node_name: str, out_node_name: str) -> TModel:
        """
        Extracts sub-model using backend-specific ModelTransformer.

        :param model_transformer: Backend-specific ModelTransformer.
        :param in_node_name: Name of the start node.
        :param out_node_name: Name of the output node.
        :return: Backend-specific sub-model.
        """
        model_extraction_command = self._backend_entity.model_extraction_command(
            [(in_node_name, 0)], [(out_node_name, 0)]
        )
        me_transformation_layout = TransformationLayout()
        me_transformation_layout.register(model_extraction_command)
        extracted_model = model_transformer.transform(me_transformation_layout)
        return extracted_model

    def _add_statistic_point(self, container: StatisticPointsContainer, point: TargetPoint, axis: int) -> None:
        """
        Adds specific statistic point.

        :param container: StatisticPointsContainer instance.
        :param point: TargetPoint for statistic collection.
        :param axis: Channel axis for the statistics calculation.
        """
        stat_collector = self._backend_entity.mean_statistic_collector(
            channel_axis=axis, num_samples=self.subset_size, inplace=self.inplace_statistics
        )
        container.add_statistic_point(
            StatisticPoint(target_point=point, tensor_collector=stat_collector, algorithm=self._algorithm_key)
        )

    def _get_bias_shift(
        self,
        model: TModel,
        input_blob: Union[TTensor, Dict[str, TTensor]],
        channel_axis: Tuple[int],
        output_fp: List[TTensor],
        output_name: str,
    ) -> TTensor:
        """
        Calculates updated bias.

        :param engine: Backend-specific engine instance for the model execution.
        :param model: Backend-specific sub-model for the execution.
        :param input_blob: Input data for the execution.
        :param channel_axis: Channel axis for the raw data aggregation.
        :param output_fp: Output data for the shift calculation.
        :param output_name: Name of the output tensor for the data collection.
        :return: Calculated bias shift.
        """
        engine = EngineFactory.create(model)
        raw_output = engine.infer(input_blob)
        q_outputs = self._backend_entity.process_model_output(raw_output, output_name)
        q_outputs = mean_per_channel(q_outputs, channel_axis)
        bias_shift = fns.stack(output_fp) - q_outputs
        return bias_shift

    def get_statistic_points(self, model: TModel, graph: NNCFGraph) -> StatisticPointsContainer:
        self._set_backend_entity(model)
        nodes_with_bias = [
            node for node in graph.get_all_nodes() if self._backend_entity.is_node_with_bias(node, graph)
        ]

        statistic_container = StatisticPointsContainer()
        for node in nodes_with_bias:
            input_port_id, output_port_id = self._backend_entity.get_activation_port_ids_for_bias_node(node)
            in_node_name, out_node_name = self._backend_entity.get_node_names_for_input_output_statistics(node, graph)

            pre_layer_statistic_point = self._backend_entity.target_point(
                TargetType.PRE_LAYER_OPERATION, in_node_name, input_port_id
            )
            post_layer_statistic_point = self._backend_entity.target_point(
                TargetType.POST_LAYER_OPERATION, out_node_name, output_port_id
            )
            channel_axis = node.metatype.output_channel_axis

            self._add_statistic_point(statistic_container, pre_layer_statistic_point, channel_axis)
            self._add_statistic_point(statistic_container, post_layer_statistic_point, channel_axis)

        return statistic_container



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional

import tensorflow as tf
from tensorflow.python.eager import context


def get_current_name_scope() -> str:
    """
    Returns current full name scope specified by `tf.name_scope(...)` method.

    :return: The name of scope.
    """
    if tf.executing_eagerly():
        return context.context().scope_name.rstrip("/")

    return tf.compat.v1.get_default_graph().get_name_scope()


def get_op_name(op_type_name: str, scope: Optional[str] = None) -> str:
    """
    Returns the name of operation from the current name of scope.

    :param op_type_name: Type name of operation.
    :param scope: The name of scope.
    :return: The name of operation.
    """
    if scope:
        if not scope.endswith("/"):
            return scope
        op_name = scope[:-1]
    else:
        current_scope = get_current_name_scope()
        if current_scope[current_scope.rfind("/") + 1 :].startswith(op_type_name.lower()):
            op_name = current_scope
        else:
            op_name = f"{current_scope}/{op_type_name}"

    # Remove `replica_*/` prefix from `op_name`.
    if op_name.startswith("replica"):
        op_name = op_name[op_name.find("/") + 1 :]

    return op_name



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from collections import Counter
from copy import deepcopy
from enum import Enum
from typing import Any, Dict, List, Optional, Set

import nncf
from nncf.common.graph import NNCFNodeName
from nncf.common.logging import nncf_logger
from nncf.common.quantization.structs import NonWeightQuantizerId
from nncf.common.quantization.structs import QuantizationScheme as QuantizationMode
from nncf.common.quantization.structs import QuantizerConfig
from nncf.common.quantization.structs import UnifiedScaleType
from nncf.common.quantization.structs import WeightQuantizerId
from nncf.common.stateful_classes_registry import CommonStatefulClassesRegistry

QuantizationPointId = int

DEFAULT_QUANTIZER_CONFIG = QuantizerConfig(
    num_bits=8, mode=QuantizationMode.SYMMETRIC, signedness_to_force=None, per_channel=False
)


class QuantizationPointType(Enum):
    WEIGHT_QUANTIZATION = 0
    ACTIVATION_QUANTIZATION = 1


class QIPointStateNames:
    TARGET_NODE_NAME = "target_node_name"


class QuantizationInsertionPointBase(ABC):
    _state_names = QIPointStateNames

    def __init__(self, target_node_name: NNCFNodeName):
        self.target_node_name = target_node_name

    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """
        return {self._state_names.TARGET_NODE_NAME: self.target_node_name}

    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> "QuantizationInsertionPointBase":
        """
        Creates the object from its state.

        :param state: Output of `get_state()` method.
        """
        return cls(**state)


@CommonStatefulClassesRegistry.register()
class WeightQuantizationInsertionPoint(QuantizationInsertionPointBase):
    def __eq__(self, other: "WeightQuantizationInsertionPoint"):
        return isinstance(other, WeightQuantizationInsertionPoint) and self.target_node_name == other.target_node_name

    def __str__(self):
        return str(WeightQuantizerId(self.target_node_name))

    def __hash__(self):
        return hash(str(self))


class AQIPointStateNames:
    INPUT_PORT_ID = "input_port_id"
    TARGET_NODE_NAME = "target_node_name"


@CommonStatefulClassesRegistry.register()
class ActivationQuantizationInsertionPoint(QuantizationInsertionPointBase):
    _state_names = AQIPointStateNames

    def __init__(self, target_node_name: NNCFNodeName, input_port_id: Optional[int] = None):
        super().__init__(target_node_name)
        self.input_port_id = input_port_id

    def __eq__(self, other: "ActivationQuantizationInsertionPoint"):
        return (
            isinstance(other, ActivationQuantizationInsertionPoint)
            and self.target_node_name == other.target_node_name
            and self.input_port_id == other.input_port_id
        )

    def __str__(self):
        return str(NonWeightQuantizerId(self.target_node_name, self.input_port_id))

    def __hash__(self):
        return hash(str(self))

    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """
        return {
            self._state_names.TARGET_NODE_NAME: self.target_node_name,
            self._state_names.INPUT_PORT_ID: self.input_port_id,
        }


class QuantizationPointBase:
    def __init__(
        self,
        quant_insertion_point: QuantizationInsertionPointBase,
        directly_quantized_operator_node_names: List[NNCFNodeName],
    ):
        self.insertion_point = quant_insertion_point
        self.directly_quantized_operator_node_names = directly_quantized_operator_node_names

    def is_activation_quantization_point(self) -> bool:
        return not self.is_weight_quantization_point()

    def is_weight_quantization_point(self) -> bool:
        return isinstance(self.insertion_point, WeightQuantizationInsertionPoint)

    def get_all_configs_list(self) -> List[QuantizerConfig]:
        raise NotImplementedError

    def __eq__(self, other):
        return self.__dict__ == other.__dict__


class SCQPointStateNames:
    QCONFIG = "qconfig"
    INSERTION_POINT = "qip"
    INSERTION_POINT_CLASS_NAME = "qip_class"
    NAMES_OF_QUANTIZED_OPS = "directly_quantized_operator_node_names"


class SingleConfigQuantizationPoint(QuantizationPointBase):
    _state_names = SCQPointStateNames

    def __init__(
        self,
        qip: QuantizationInsertionPointBase,
        qconfig: QuantizerConfig,
        directly_quantized_operator_node_names: List[NNCFNodeName],
    ):
        super().__init__(qip, directly_quantized_operator_node_names)
        self.qconfig = deepcopy(qconfig)

    def __str__(self):
        return str(self.insertion_point) + " " + str(self.qconfig)

    def get_all_configs_list(self) -> List[QuantizerConfig]:
        return [self.qconfig]

    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """
        return {
            self._state_names.INSERTION_POINT: self.insertion_point.get_state(),
            self._state_names.INSERTION_POINT_CLASS_NAME: self.insertion_point.__class__.__name__,
            self._state_names.QCONFIG: self.qconfig.get_state(),
            self._state_names.NAMES_OF_QUANTIZED_OPS: self.directly_quantized_operator_node_names,
        }

    @classmethod
    def from_state(cls, state: Dict[str, Any]) -> "SingleConfigQuantizationPoint":
        """
        Creates the object from its state.

        :param state: Output of `get_state()` method.
        """
        insertion_point_cls_name = state[cls._state_names.INSERTION_POINT_CLASS_NAME]
        insertion_point_cls = CommonStatefulClassesRegistry.get_registered_class(insertion_point_cls_name)
        insertion_point = insertion_point_cls.from_state(state[cls._state_names.INSERTION_POINT])
        kwargs = {
            cls._state_names.INSERTION_POINT: insertion_point,
            cls._state_names.QCONFIG: QuantizerConfig.from_state(state[cls._state_names.QCONFIG]),
            cls._state_names.NAMES_OF_QUANTIZED_OPS: state[cls._state_names.NAMES_OF_QUANTIZED_OPS],
        }
        return cls(**kwargs)


class MultiConfigQuantizationPoint(QuantizationPointBase):
    def __init__(
        self,
        qip: QuantizationInsertionPointBase,
        possible_qconfigs: List[QuantizerConfig],
        directly_quantized_operator_node_names: List[NNCFNodeName],
    ):
        super().__init__(qip, directly_quantized_operator_node_names)
        self.possible_qconfigs = possible_qconfigs

    @property
    def possible_qconfigs(self):
        return deepcopy(self._possible_qconfigs)

    @possible_qconfigs.setter
    def possible_qconfigs(self, qconfigs: List[QuantizerConfig]):
        self._possible_qconfigs = deepcopy(qconfigs)

    def select_qconfig(self, qconfig: QuantizerConfig) -> SingleConfigQuantizationPoint:
        if qconfig not in self.possible_qconfigs:
            # Allow selecting an "unsigned" or "signed" version if "any-signed" version is present
            qconfig_any = deepcopy(qconfig)
            qconfig_any.signedness_to_force = None
            if qconfig_any not in self.possible_qconfigs:
                raise ValueError(
                    "Invalid selection for a quantizer config - "
                    "tried to select {} among [{}]".format(qconfig, ",".join([str(q) for q in self.possible_qconfigs]))
                )
            qconfig = qconfig_any
        return SingleConfigQuantizationPoint(self.insertion_point, qconfig, self.directly_quantized_operator_node_names)

    def __str__(self):
        return str(self.insertion_point) + " " + ";".join([str(qc) for qc in self.possible_qconfigs])

    def get_all_configs_list(self) -> List[QuantizerConfig]:
        return self.possible_qconfigs


class QuantizerSetupBase:
    def __init__(self):
        self.quantization_points: Dict[QuantizationPointId, QuantizationPointBase] = {}
        self.unified_scale_groups: Dict[int, Set[QuantizationPointId]] = {}
        self.shared_input_operation_set_groups: Dict[int, Set[QuantizationPointId]] = {}
        self._next_unified_scale_gid = 0
        self._next_shared_inputs_gid = 0

    def add_independent_quantization_point(self, qp: QuantizationPointBase):
        if self.quantization_points.keys():
            new_id = max(self.quantization_points.keys()) + 1
        else:
            new_id = 0
        self.quantization_points[new_id] = qp

    def register_unified_scale_group(self, qp_group: List[QuantizationPointId]) -> int:
        for qp_id in qp_group:
            gid = self.get_unified_scale_group_id(qp_id) is not None
            if gid:
                raise nncf.InternalError("QP id {} is already in unified scale group {}".format(qp_id, gid))
        gid = self._next_unified_scale_gid
        self.unified_scale_groups[self._next_unified_scale_gid] = set(qp_group)
        self._next_unified_scale_gid += 1
        return gid

    def register_shared_inputs_group(self, qp_group: List[QuantizationPointId]) -> int:
        for qp_id in qp_group:
            gid = self.get_shared_inputs_group_id(qp_id) is not None
            if gid:
                raise nncf.InternalError("QP id {} is already in shared input group {}".format(qp_id, gid))
        gid = self._next_shared_inputs_gid
        self.shared_input_operation_set_groups[self._next_shared_inputs_gid] = set(qp_group)
        self._next_shared_inputs_gid += 1
        return gid

    def __discard_independent(self, id_: QuantizationPointId):
        if id_ in self.quantization_points:
            self.quantization_points.pop(id_)
        for unified_scale_group in self.unified_scale_groups.values():
            unified_scale_group.discard(id_)

    def discard(self, id_: QuantizationPointId, keep_shared_input_qps: bool = False):
        if id_ in self.quantization_points:
            self.__discard_independent(id_)

            # If a quantizer from the shared operation set group is removed, then the
            # entire group has to be removed from the setup, otherwise an operation would have a mix between
            # quantized and unquantized inputs.
            indices_to_delete = []
            for gid, shared_input_operation_set_group in self.shared_input_operation_set_groups.items():
                if id_ in shared_input_operation_set_group:
                    shared_input_operation_set_group.discard(id_)
                    indices_to_delete.append(gid)

            if not keep_shared_input_qps:
                for idx in sorted(indices_to_delete, reverse=True):
                    for additional_id in self.shared_input_operation_set_groups[idx]:
                        self.__discard_independent(additional_id)
                    del self.shared_input_operation_set_groups[idx]

    def get_unified_scale_group_id(self, qp_id: QuantizationPointId) -> Optional[int]:
        for gid, unified_scale_group in self.unified_scale_groups.items():
            if qp_id in unified_scale_group:
                return gid
        return None

    def get_shared_inputs_group_id(self, qp_id: QuantizationPointId) -> Optional[int]:
        for gid, shared_inputs_group in self.shared_input_operation_set_groups.items():
            if qp_id in shared_inputs_group:
                return gid
        return None

    def register_existing_qp_id_in_unified_scale_group(self, qp_id: QuantizationPointId, unified_scale_gid: int):
        gid = self.get_unified_scale_group_id(qp_id)
        if gid is not None:
            raise nncf.InternalError("QP id {} is already in unified scale group {}".format(qp_id, gid))
        self.unified_scale_groups[unified_scale_gid].add(qp_id)

    def register_existing_qp_id_in_shared_input_group(self, qp_id: QuantizationPointId, shared_inputs_gid: int):
        gid = self.get_shared_inputs_group_id(qp_id)
        if gid is not None:
            raise nncf.InternalError("QP id {} is already in shared inputs group {}".format(qp_id, gid))
        self.shared_input_operation_set_groups[shared_inputs_gid].add(qp_id)

    def remove_unified_scale_from_point(self, qp_id: QuantizationPointId):
        gid = self.get_unified_scale_group_id(qp_id)
        if gid is None:
            nncf_logger.debug(
                f"Attempted to remove QP id {qp_id} from associated unified scale group, but the QP"
                f"is not in any unified scale group - ignoring."
            )
            return
        self.unified_scale_groups[gid].discard(qp_id)
        if not self.unified_scale_groups[gid]:
            nncf_logger.debug(f"Removed last entry from a unified scale group {gid} - removing group itself")
            self.unified_scale_groups.pop(gid)

    def equivalent_to(self, other: "QuantizerSetupBase") -> bool:
        this_qp_id_to_other_qp_id_dict: Dict[QuantizationPointId, QuantizationPointId] = {}

        def _compare_qps(first: "QuantizerSetupBase", second: "QuantizerSetupBase") -> bool:
            for this_qp_id, this_qp in first.quantization_points.items():
                matches: List[QuantizationPointId] = []
                for other_qp_id, other_qp in second.quantization_points.items():
                    if this_qp == other_qp:
                        matches.append(other_qp_id)
                if len(matches) == 0:
                    return False
                assert len(matches) == 1  # separate quantization points should not compare equal to each other
                this_qp_id_to_other_qp_id_dict[this_qp_id] = matches[0]
            return True

        def _compare_shared_input_groups(first: "QuantizerSetupBase", second: "QuantizerSetupBase") -> bool:
            for this_same_input_group_set in first.shared_input_operation_set_groups.values():
                translated_id_set = set(
                    this_qp_id_to_other_qp_id_dict[this_qp_id] for this_qp_id in this_same_input_group_set
                )
                matches = []

                for other_shared_inputs_group in second.shared_input_operation_set_groups.values():
                    if translated_id_set == other_shared_inputs_group:
                        matches.append(other_shared_inputs_group)
                if not matches:
                    return False
                assert len(matches) == 1  # shared inputs group entries should be present in only one group
            return True

        def _compare_unified_scale_groups(first: "QuantizerSetupBase", second: "QuantizerSetupBase") -> bool:
            for this_unified_scales_group in first.unified_scale_groups.values():
                translated_id_set = set(
                    this_qp_id_to_other_qp_id_dict[this_qp_id] for this_qp_id in this_unified_scales_group
                )
                matches = []
                for other_unified_scales_group in second.unified_scale_groups.values():
                    if translated_id_set == other_unified_scales_group:
                        matches.append(other_unified_scales_group)
                if not matches:
                    return False
                assert len(matches) == 1  # unified scale group entries should be present in only one group
            return True

        return (
            _compare_qps(self, other)
            and _compare_qps(other, self)
            and _compare_shared_input_groups(self, other)
            and _compare_shared_input_groups(self, other)
            and _compare_unified_scale_groups(self, other)
            and _compare_unified_scale_groups(self, other)
        )


class SCQSetupStateNames:
    SHARED_INPUT_OPERATION_SET_GROUPS = "shared_input_operation_set_groups"
    UNIFIED_SCALE_GROUPS = "unified_scale_groups"
    QUANTIZATION_POINTS = "quantization_points"


class SingleConfigQuantizerSetup(QuantizerSetupBase):
    _state_names = SCQSetupStateNames

    def __init__(self):
        super().__init__()
        self.quantization_points: Dict[QuantizationPointId, SingleConfigQuantizationPoint] = {}

    def get_state(self) -> Dict:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """

        def set2list(pair):
            i, qp_id_set = pair
            return i, list(qp_id_set)

        quantization_points_state = {qp_id: qp.get_state() for qp_id, qp in self.quantization_points.items()}
        unified_scale_groups_state = dict(map(set2list, self.unified_scale_groups.items()))
        shared_input_operation_set_groups_state = dict(map(set2list, self.shared_input_operation_set_groups.items()))
        return {
            self._state_names.QUANTIZATION_POINTS: quantization_points_state,
            self._state_names.UNIFIED_SCALE_GROUPS: unified_scale_groups_state,
            self._state_names.SHARED_INPUT_OPERATION_SET_GROUPS: shared_input_operation_set_groups_state,
        }

    @classmethod
    def from_state(cls, state: Dict) -> "SingleConfigQuantizerSetup":
        """
        Creates the object from its state.

        :param state: Output of `get_state()` method.
        """
        setup = SingleConfigQuantizerSetup()

        def decode_qp(pair):
            str_qp_id, qp_state = pair
            return int(str_qp_id), SingleConfigQuantizationPoint.from_state(qp_state)

        def list2set(pair):
            str_idx, qp_id_list = pair
            return int(str_idx), set(qp_id_list)

        setup.quantization_points = dict(map(decode_qp, state[cls._state_names.QUANTIZATION_POINTS].items()))
        setup.unified_scale_groups = dict(map(list2set, state[cls._state_names.UNIFIED_SCALE_GROUPS].items()))
        shared_input_operation_set_groups_state = state[cls._state_names.SHARED_INPUT_OPERATION_SET_GROUPS]
        setup.shared_input_operation_set_groups = dict(map(list2set, shared_input_operation_set_groups_state.items()))
        return setup


class MultiConfigQuantizerSetup(QuantizerSetupBase):
    def __init__(self):
        super().__init__()
        self.quantization_points: Dict[QuantizationPointId, MultiConfigQuantizationPoint] = {}
        self._unified_scale_qpid_vs_type: Dict[QuantizationPointId, UnifiedScaleType] = {}

    def register_unified_scale_group_with_types(
        self, qp_group: List[QuantizationPointId], us_types: List[UnifiedScaleType]
    ) -> int:
        assert len(qp_group) == len(us_types)
        gid = super().register_unified_scale_group(qp_group)
        for qp_id, us_type in zip(qp_group, us_types):
            self._unified_scale_qpid_vs_type[qp_id] = us_type
        return gid

    def select_qconfigs(
        self, qp_id_vs_selected_qconfig_dict: Dict[QuantizationPointId, QuantizerConfig], strict: bool = True
    ) -> SingleConfigQuantizerSetup:
        retval = SingleConfigQuantizerSetup()
        retval.unified_scale_groups = deepcopy(self.unified_scale_groups)
        retval.shared_input_operation_set_groups = deepcopy(self.shared_input_operation_set_groups)

        if Counter(qp_id_vs_selected_qconfig_dict.keys()) != Counter(self.quantization_points.keys()):
            raise ValueError(
                "The set of quantization points for a selection is inconsistent with quantization"
                "points in the quantizer setup!"
            )
        for qp_id, qp in self.quantization_points.items():
            if strict:
                retval.quantization_points[qp_id] = qp.select_qconfig(qp_id_vs_selected_qconfig_dict[qp_id])
            else:
                multi_qp = qp
                qconfig = qp_id_vs_selected_qconfig_dict[qp_id]
                retval.quantization_points[qp_id] = SingleConfigQuantizationPoint(
                    multi_qp.insertion_point, qconfig, multi_qp.directly_quantized_operator_node_names
                )

        # Segregate the unified scale groups into sub-groups based on what exact config was chosen.
        for us_group in self.unified_scale_groups.values():
            per_channel_qids = set()
            per_tensor_qids = set()
            for us_qid in us_group:
                final_qconfig = retval.quantization_points[us_qid].qconfig
                if final_qconfig.per_channel:
                    per_channel_qids.add(us_qid)
                else:
                    per_tensor_qids.add(us_qid)

            if per_tensor_qids:
                for qid in per_tensor_qids:
                    retval.remove_unified_scale_from_point(qid)

                retval.register_unified_scale_group(list(per_tensor_qids))

            for per_channel_qid in per_channel_qids:
                us_type = self._unified_scale_qpid_vs_type[per_channel_qid]
                if us_type is UnifiedScaleType.UNIFY_ONLY_PER_TENSOR:
                    nncf_logger.debug(
                        "Per-channel quantizer config selected in a MultiConfigQuantizerSetup for a "
                        "unified scale point that only supports per-tensor scale unification, disabling "
                        "unified scales for this point."
                    )
                retval.remove_unified_scale_from_point(per_channel_qid)

        return retval

    def select_first_qconfig_for_each_point(self) -> SingleConfigQuantizerSetup:
        qp_id_vs_qconfig_dict: Dict[QuantizationPointId, QuantizerConfig] = {}
        for qp_id, qp in self.quantization_points.items():
            qp_id_vs_qconfig_dict[qp_id] = qp.possible_qconfigs[0]
        return self.select_qconfigs(qp_id_vs_qconfig_dict)

    @classmethod
    def from_single_config_setup(cls, single_conf_setup: SingleConfigQuantizerSetup) -> "MultiConfigQuantizerSetup":
        retval = cls()
        for qp_id, qp in single_conf_setup.quantization_points.items():
            multi_pt = MultiConfigQuantizationPoint(
                qip=qp.insertion_point,
                possible_qconfigs=[deepcopy(qp.qconfig)],
                directly_quantized_operator_node_names=qp.directly_quantized_operator_node_names,
            )
            retval.quantization_points[qp_id] = multi_pt
        for qp_set in single_conf_setup.unified_scale_groups.values():
            qp_list = list(qp_set)
            qp_types = [UnifiedScaleType.UNIFY_ALWAYS for _ in qp_list]
            retval.register_unified_scale_group_with_types(qp_list, qp_types)
        for qp_set in single_conf_setup.shared_input_operation_set_groups.values():
            qp_list = list(qp_set)
            retval.register_shared_inputs_group(qp_list)
        return retval



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from nncf.config.schemata.basic import BOOLEAN
from nncf.config.schemata.basic import make_string_or_array_of_strings_schema
from nncf.config.schemata.basic import with_attributes
from nncf.config.schemata.common.initialization import BATCHNORM_ADAPTATION_SCHEMA
from nncf.config.schemata.defaults import VALIDATE_SCOPES

IGNORED_SCOPES_DESCRIPTION = (
    "A list of model control flow graph node scopes to be ignored for this "
    "operation - functions as an 'allowlist'. Optional."
)
TARGET_SCOPES_DESCRIPTION = (
    "A list of model control flow graph node scopes to be considered for this operation"
    " - functions as a 'denylist'. Optional."
)
VALIDATE_SCOPES_DESCRIPTION = (
    "If set to True, then a RuntimeError will be raised if the names of the "
    "ignored/target scopes do not match the names of the scopes in the model graph."
)
SCOPING_PROPERTIES = {
    "ignored_scopes": with_attributes(
        make_string_or_array_of_strings_schema(),
        description=IGNORED_SCOPES_DESCRIPTION,
        examples=["{re}conv.*", ["LeNet/relu_0", "LeNet/relu_1"]],
    ),
    "target_scopes": with_attributes(
        make_string_or_array_of_strings_schema(),
        description=TARGET_SCOPES_DESCRIPTION,
        examples=[
            [
                "UNet/ModuleList[down_path]/UNetConvBlock[1]/Sequential[block]/Conv2d[0]",
                "UNet/ModuleList[down_path]/UNetConvBlock[2]/Sequential[block]/Conv2d[0]",
                "UNet/ModuleList[down_path]/UNetConvBlock[3]/Sequential[block]/Conv2d[0]",
                "UNet/ModuleList[down_path]/UNetConvBlock[4]/Sequential[block]/Conv2d[0]",
            ],
            "UNet/ModuleList\\[up_path\\].*",
        ],
    ),
    "validate_scopes": with_attributes(
        BOOLEAN,
        description=VALIDATE_SCOPES_DESCRIPTION,
        default=VALIDATE_SCOPES,
    ),
}
GENERIC_INITIALIZER_SCHEMA = {
    "type": "object",
    "properties": {"batchnorm_adaptation": BATCHNORM_ADAPTATION_SCHEMA},
    "additionalProperties": False,
}



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Callable, Dict, List, NoReturn

from nncf.api.compression import CompressionLoss
from nncf.api.compression import CompressionScheduler
from nncf.api.compression import CompressionStage
from nncf.common.initialization.batchnorm_adaptation import BatchnormAdaptationAlgorithm
from nncf.common.logging import nncf_logger
from nncf.common.statistics import NNCFStatistics
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.elasticity_controller import ElasticityController
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.elasticity_dim import ElasticityDim
from nncf.experimental.torch.nas.bootstrapNAS.elasticity.multi_elasticity_handler import MultiElasticityHandler
from nncf.experimental.torch.nas.bootstrapNAS.training.base_training import BNASTrainingController
from nncf.experimental.torch.nas.bootstrapNAS.training.lr_scheduler import GlobalLRScheduler
from nncf.experimental.torch.nas.bootstrapNAS.training.lr_scheduler import StageLRScheduler
from nncf.experimental.torch.nas.bootstrapNAS.training.scheduler import BootstrapNASScheduler
from nncf.experimental.torch.nas.bootstrapNAS.training.scheduler import NASSchedulerParams
from nncf.experimental.torch.nas.bootstrapNAS.training.stage_descriptor import StageDescriptor
from nncf.torch.nncf_network import NNCFNetwork


class PSControllerStateNames:
    ELASTICITY_CONTROLLER_STATE = "elasticity_controller_compression_state"
    LR_GLOBAL_SCHEDULE_STATE = "learning_rate_global_schedule_state"


class ProgressiveShrinkingController(BNASTrainingController):
    """
    Serves as a handle to the additional modules, parameters and hooks inserted
    into the original uncompressed model in order to train a supernet using Progressive Shrinking procedure
    from OFA (https://arxiv.org/abs/1908.09791).
    Hosts entities that are to be used during the training process, such as compression scheduler and
    compression loss.
    """

    _ps_state_names = PSControllerStateNames

    def __init__(
        self,
        target_model: NNCFNetwork,
        elasticity_ctrl: ElasticityController,
        bn_adaptation: BatchnormAdaptationAlgorithm,
        progressivity_of_elasticity: List[ElasticityDim],
        schedule_params: NASSchedulerParams,
        lr_schedule_config: Dict[str, Any],
        compression_loss_func: Callable,
    ):
        super().__init__(target_model)
        self._elasticity_ctrl = elasticity_ctrl
        self._bn_adaptation = bn_adaptation
        self._progressivity_of_elasticity = progressivity_of_elasticity
        self._target_model = target_model
        self._loss = compression_loss_func
        self._available_elasticity_dims = self.multi_elasticity_handler.get_available_elasticity_dims()
        self._lr_schedule_config = lr_schedule_config
        self._scheduler = BootstrapNASScheduler(
            self, schedule_params, self._available_elasticity_dims, self._progressivity_of_elasticity
        )
        self._sample_rate = 1

    def set_training_lr_scheduler_args(self, optimizer, train_iters):
        params = self._lr_schedule_config.get("params", {})
        num_epochs = params.get("num_epochs", None)
        base_lr = params.get("base_lr", None)

        if base_lr is not None:
            nncf_logger.info("Global LR scheduler in use")
            # Global lr scheduler
            if num_epochs is None:
                params["num_epochs"] = self.get_total_num_epochs()
            lr_scheduler = GlobalLRScheduler(optimizer, train_iters, **params)
        else:
            nncf_logger.info("Stage LR scheduler in use")
            lr_scheduler = StageLRScheduler(optimizer, train_iters)
        self._scheduler.lr_scheduler = lr_scheduler

    @property
    def lr_schedule_config(self) -> str:
        """
        Gets access to learning rate scheduler configuration.

        :return: learning rate scheduler
        """
        return self._lr_schedule_config

    @lr_schedule_config.setter
    def lr_schedule_config(self, val: Dict[str, Any]) -> NoReturn:
        self._lr_schedule_config = val

    @property
    def multi_elasticity_handler(self) -> MultiElasticityHandler:
        """
        Gets access to multi elasticity handler to perform some actions with supernet or subnets.

        :return: multi elasticity handler
        """
        return self._elasticity_ctrl.multi_elasticity_handler

    @property
    def elasticity_controller(self) -> ElasticityController:
        """
        Gets access to elasticity controller. Usually it's needed for saving its state for further resuming in the
        search part.

        :return: elasticity controller
        """
        return self._elasticity_ctrl

    @property
    def loss(self) -> CompressionLoss:
        """
        :return: The instance of the `CompressionLoss`.
        """
        return self._loss

    @property
    def scheduler(self) -> CompressionScheduler:
        """
        :return: The instance of the `CompressionScheduler`.
        """
        return self._scheduler

    def step(self) -> None:
        """
        Should be called at the beginning of each training step for activation some Subnet(s).
        """
        if self._scheduler.current_step % self._sample_rate == 0:
            self.multi_elasticity_handler.activate_random_subnet()
            nncf_logger.debug(f"Active config: {self.multi_elasticity_handler.get_active_config()}")

    def prepare_for_validation(self) -> None:
        """
        Performs some action on active subnet or supernet before validation. For instance, it can be the batchnorm
        adaptation to achieve the best accuracy on validation.
        """
        if self._bn_adaptation:
            self._run_batchnorm_adaptation(self._target_model)

    def get_total_num_epochs(self) -> int:
        """
        Returns total number of epochs required for the supernet training.

        :return: number of epochs
        """
        return self._scheduler.get_total_training_epochs()

    def set_stage(self, stage_desc: StageDescriptor) -> None:
        """
        Set a new training stage with parameters from a given stage descriptor

        :param stage_desc: describes parameters of the training stage that should be enabled
        """
        for elasticity_dim in self._available_elasticity_dims:
            if elasticity_dim in stage_desc.train_dims:
                self.multi_elasticity_handler.enable_elasticity(elasticity_dim)
            else:
                self.multi_elasticity_handler.disable_elasticity(elasticity_dim)

        width_handler = self.multi_elasticity_handler.width_handler
        depth_handler = self.multi_elasticity_handler.depth_handler
        if width_handler is not None:
            if stage_desc.reorg_weights:
                width_handler.reorganize_weights()
            width_indicator = stage_desc.width_indicator
            if width_indicator:
                width_handler.width_num_params_indicator = width_indicator

        depth_indicator = stage_desc.depth_indicator
        if depth_handler and depth_indicator:
            depth_handler.depth_indicator = depth_indicator

        if stage_desc.bn_adapt:
            self._run_batchnorm_adaptation(self._target_model)

        self._sample_rate = stage_desc.sample_rate

    def statistics(self, quickly_collected_only: bool = False) -> NNCFStatistics:
        """
        Returns a `Statistics` class instance that contains compression algorithm statistics.

        :param quickly_collected_only: Enables collection of the statistics that
            don't take too much time to compute. Can be helpful for the case when
            need to keep track of statistics on each training batch/step/iteration.
        :return: A `Statistics` class instance that contains compression algorithm statistics.
        """
        return NNCFStatistics()

    def compression_stage(self) -> CompressionStage:
        """
        Returns the compression stage. Should be used on saving best checkpoints
        to distinguish between uncompressed, partially compressed, and fully
        compressed models.

        :return: The compression stage of the target model.
        """
        if self._scheduler.is_final_stage():
            return CompressionStage.FULLY_COMPRESSED
        return CompressionStage.PARTIALLY_COMPRESSED

    def load_state(self, state: Dict[str, Dict[str, Any]]) -> None:
        """
        Loads the compression controller state from the map of algorithm name to the dictionary with state attributes.

        :param state: map of the algorithm name to the dictionary with the corresponding state attributes.
        """
        self._lr_schedule_config = state[self._ps_state_names.LR_GLOBAL_SCHEDULE_STATE]
        super().load_state(state)
        elasticity_ctrl_state = state[self._ps_state_names.ELASTICITY_CONTROLLER_STATE]
        self._elasticity_ctrl.load_state(elasticity_ctrl_state)

    def get_state(self) -> Dict[str, Dict[str, Any]]:
        """
        Returns compression controller state, which is the map of the algorithm name to the dictionary with the
        corresponding state attributes.

        :return: The compression controller state.
        """
        state = super().get_state()
        state[self._ps_state_names.ELASTICITY_CONTROLLER_STATE] = self._elasticity_ctrl.get_state()
        state[self._ps_state_names.LR_GLOBAL_SCHEDULE_STATE] = self._lr_schedule_config
        return state

    def _run_batchnorm_adaptation(self, model):
        if self._bn_adaptation is None:
            nncf_logger.warning("Batchnorm adaptation requested but it hasn't been enabled for training.")
        self._bn_adaptation.run(model)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import abstractmethod
from typing import Dict, List, Optional, TypeVar

from nncf.common.graph import NNCFGraph
from nncf.common.logging.track_progress import track
from nncf.common.utils.registry import Registry
from nncf.experimental.tensor import Tensor
from nncf.experimental.tensor import functions as fns
from nncf.experimental.tensor.definitions import TensorDataType
from nncf.parameters import SensitivityMetric
from nncf.quantization.algorithms.weight_compression.backend import WeightCompressionAlgoBackend
from nncf.quantization.algorithms.weight_compression.config import WeightCompressionConfig
from nncf.quantization.algorithms.weight_compression.config import WeightCompressionParameters
from nncf.quantization.algorithms.weight_compression.weight_lowering import do_integer_quantization
from nncf.quantization.algorithms.weight_compression.weight_lowering import get_integer_quantization_error

TModel = TypeVar("TModel")
MIXED_PRECISION_CRITERIA = Registry("mixed_precision_criteria")
THE_LOWEST_SENSITIVITY = 0


class MixedPrecisionCriterion:
    """
    Assigns mixed quantization scheme (e.g. uniform int8 or non-uniform nf4) for weights based on some criteria.
    """

    def __init__(
        self,
        model: TModel,
        graph: NNCFGraph,
        backend_entity: WeightCompressionAlgoBackend,
        weight_params: List[WeightCompressionParameters],
        primary_config: WeightCompressionConfig,
        ratio: float,
        activations: Optional[Dict[str, List[Tensor]]] = None,
    ):
        """
        :param model: The model.
        :param graph: The model graph associated with the model.
        :param backend_entity: The instance of the WeightCompressionAlgoBackend.
        :param weight_params: Weight compression parameters which determines how and what weight should be compressed.
        :param primary_config: Configuration on how to compress (quantize) weights to primary precision.
        :param ratio: The ratio between primary and backup precisions (e.g. 0.9 means 90% of layers quantized to NF4
            and the rest to INT8_ASYM).
        :param activations: The input activations of the nodes to be quantized.
        """
        self._model = model
        self._graph = graph
        self._backend_entity = backend_entity
        self._weight_params = weight_params
        self._activations = activations
        self._primary_config = primary_config
        self._ratio = ratio

    @abstractmethod
    def _calc_sensitivity(self) -> List[float]:
        """
        Calculates sensitivity of each layer according to a criterion.

        :return: List of values per node to be quantized.
        """

    def assign_mixed_precision(self) -> None:
        """
        Assigns quantization precision based on computed layers' sensitivities, ratio of parameters.
        """
        scores = self._calc_sensitivity()
        num_all_weights = sum(wp.num_weights for wp in self._weight_params)

        indexes_of_layers_in_ascending_order_of_scores = [
            i[0] for i in sorted(enumerate(scores), reverse=False, key=lambda x: x[1])
        ]
        num_weights_in_4bit = 0
        for index in indexes_of_layers_in_ascending_order_of_scores:
            weight_param = self._weight_params[index]
            current_ratio = (num_weights_in_4bit + weight_param.num_weights) / num_all_weights
            if current_ratio >= self._ratio:
                break
            weight_param.compression_config = self._primary_config
            num_weights_in_4bit += weight_param.num_weights


@MIXED_PRECISION_CRITERIA.register(SensitivityMetric.WEIGHT_QUANTIZATION_ERROR)
class DataFreeCriterion(MixedPrecisionCriterion):
    """
    A baseline mixed precision criterion that is based on quantization noise of weights only.
    """

    def _calc_weight_sensitivity(self, weight_param: WeightCompressionParameters) -> float:
        weight = self._backend_entity.get_weight(
            weight_param.node_with_weight, weight_param.weight_port_id, self._model, self._graph
        )
        backup_config = weight_param.compression_config
        reduction_axes = weight_param.reduction_axes
        int_error = get_integer_quantization_error(weight, reduction_axes, backup_config)
        eps = fns.finfo(weight).eps
        return 1 / (int_error + eps)

    def _calc_score_per_node(self, weight_param: WeightCompressionParameters) -> float:
        weight_score = self._calc_weight_sensitivity(weight_param)
        return weight_score

    def _calc_sensitivity(self) -> List[float]:
        scores = []
        for weight_param in track(self._weight_params, description="Mixed-Precision assignment"):
            scores.append(self._calc_score_per_node(weight_param))
        return scores


class DataBasedCriterion(DataFreeCriterion):
    """
    Data-based mixed precision criterion that takes into account outliers in the input activations.
    Expecting activations of the following shape: [seq_length, hidden_dim]
    """

    @staticmethod
    @abstractmethod
    def _calc_activation_sensitivity(activations: List[Tensor]) -> float:
        pass

    def _calc_score_per_node(self, weight_param: WeightCompressionParameters):
        """
        NOTE: Data-based criteria for assigning 4-bit/8-bit precisions are valid for Matmul operations only.
        However, in some cases it can be beneficial to quantize Gather layers to 4-bit.
        Since there's no data-aware estimation of sensitivity in these layers, they receive the lowest sensitivity.
        It allows assigning Gather operation 4-bit in the first place.
        """
        if weight_param.node_with_weight.metatype in self._backend_entity.embedding_metatypes:
            return THE_LOWEST_SENSITIVITY
        weight_score = self._calc_weight_sensitivity(weight_param)
        activation_score = self._calc_activation_sensitivity(self._activations[weight_param.node_with_weight.node_name])
        return weight_score * activation_score


@MIXED_PRECISION_CRITERIA.register(SensitivityMetric.HESSIAN_INPUT_ACTIVATION)
class HAWQCriterion(DataBasedCriterion):
    """
    Calculates the average Hessian trace of weights with respect to the layer-wise quantization error
    multiplied by L2 norm of 8-bit quantization noise.
    """

    @staticmethod
    def _calc_activation_sensitivity(activations: List[Tensor]) -> float:
        htrace = 0
        nsamples = len(activations)
        for inp in activations:
            # NOTE: average trace?? divide by number of diagonal elements
            htrace += fns.sum(fns.multiply(inp, inp)).item()
            # normalize by sequence_length - the same for all activations
            # normalize by hidden dimension
            htrace /= inp.size
        htrace *= 2 / nsamples
        return htrace

    def _calc_weight_sensitivity(self, weight_param: WeightCompressionParameters) -> float:
        weight = self._backend_entity.get_weight(
            weight_param.node_with_weight, weight_param.weight_port_id, self._model, self._graph
        )
        backup_config = weight_param.compression_config
        reduction_axes = weight_param.reduction_axes

        orig_shape = weight.shape

        if weight.dtype != TensorDataType.float32:
            weight = weight.astype(TensorDataType.float32)

        compressed_weights, scale, zero_point = do_integer_quantization(weight, reduction_axes, backup_config)
        decompressed_weight = (compressed_weights - zero_point).astype(weight.dtype) * scale
        decompressed_weight = decompressed_weight.reshape(orig_shape)
        return fns.linalg.norm(decompressed_weight - weight, ord="fro").item()


@MIXED_PRECISION_CRITERIA.register(SensitivityMetric.MEAN_ACTIVATION_VARIANCE)
class MeanVarianceCriterion(DataBasedCriterion):
    """
    The mean variance of the layers' inputs multiplied by inverted 8-bit quantization noise.
    """

    @staticmethod
    def _calc_activation_sensitivity(activations: List[Tensor]) -> float:
        return fns.mean(fns.stack([fns.mean(fns.var(inp, axis=0)) for inp in activations])).item()


@MIXED_PRECISION_CRITERIA.register(SensitivityMetric.MAX_ACTIVATION_VARIANCE)
class MaxVarianceCriterion(DataBasedCriterion):
    """
    The maximum variance of the layers' inputs multiplied by inverted 8-bit quantization noise.
    """

    @staticmethod
    def _calc_activation_sensitivity(activations: List[Tensor]) -> float:
        return fns.mean(fns.stack([fns.max(fns.var(inp, axis=0)) for inp in activations])).item()


@MIXED_PRECISION_CRITERIA.register(SensitivityMetric.MEAN_ACTIVATION_MAGNITUDE)
class MeanMaxCriterion(DataBasedCriterion):
    """
    The mean magnitude of the layers' inputs multiplied by inverted 8-bit quantization noise.
    """

    @staticmethod
    def _calc_activation_sensitivity(activations: List[Tensor]) -> float:
        return fns.mean(fns.stack([fns.mean(fns.max(fns.abs(inp), axis=0)) for inp in activations])).item()



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from collections import defaultdict
from typing import Any, Callable, Dict, Optional, Set

import torch

from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph.definitions import MODEL_CONST_OP_NAME
from nncf.common.graph.operator_metatypes import CONST_NOOP_METATYPES
from nncf.common.graph.operator_metatypes import INPUT_NOOP_METATYPES
from nncf.torch.dynamic_graph.context import TracingContext
from nncf.torch.dynamic_graph.graph import DynamicGraph
from nncf.torch.dynamic_graph.graph_tracer import GraphTracer
from nncf.torch.dynamic_graph.layer_attributes_handlers import set_nodes_attributes_in_nncf_graph
from nncf.torch.dynamic_graph.scope import Scope
from nncf.torch.graph.graph import PTNNCFGraph
from nncf.torch.graph.operator_metatypes import PT_OPERATOR_METATYPES
from nncf.torch.graph.operator_metatypes import QUANTIZE_NODE_TYPES


class GraphBuilder:
    def __init__(self, custom_forward_fn: Callable[[torch.nn.Module], Any]):
        self.custom_forward_fn = custom_forward_fn

    def build_dynamic_graph(
        self,
        model: torch.nn.Module,
        context_to_use: Optional[TracingContext] = None,
        as_eval: bool = False,
        trace_parameters: bool = False,
    ) -> DynamicGraph:
        """
        Builds DynamicGraph from the given Torch model.

        :param model: Model to build DynamicGraph from.
        :param context_to_use: Tracing context to use during the DynamicGraph building. Creates new tracing context
            if context to use is not specified.
        :param as_eval: Should given model be switched to eval mode before the graph tracing or not. Default is False.
        :param trace_parameters: Whether trace model parameters during the DynamicGraph building or not.
            Default is False.
        :return: DynamicGraph constructed from given model.
        """
        tracer = GraphTracer(self.custom_forward_fn)
        return tracer.trace_graph(model, context_to_use, as_eval, trace_parameters)

    def build_graph(
        self,
        model: torch.nn.Module,
        context_to_use: Optional[TracingContext] = None,
        as_eval: bool = False,
        trace_parameters: bool = False,
    ) -> PTNNCFGraph:
        """
        Builds PTNNCFGraph representation from the given Torch model.

        :param model: Model to build PTNNCFGraph from.
        :param context_to_use: Tracing context to use during the PTNNCFGraph building. Creates new tracing context
            if context to use is not specified.
        :param as_eval: Should given model be switched to eval mode before the graph tracing or not. Default is False.
        :param trace_parameters: Whether trace model parameters during the PTNNCFGraph building or not.
            Default is False.
        :return: PTNNCFGraph constructed from given model.
        """
        dynamic_graph = self.build_dynamic_graph(model, context_to_use, as_eval, trace_parameters)
        return GraphConverter.convert(dynamic_graph, trace_parameters)


class GraphConverter:
    @staticmethod
    def convert(dynamic_graph: DynamicGraph, traced_parameters) -> PTNNCFGraph:
        module_id_vs_known_op_addrs_map: Dict[int, Set[Scope]] = defaultdict(set)
        for dynamic_graph_node in dynamic_graph.get_all_nodes():
            # Skip const nodes to detect shared nodes
            if dynamic_graph_node.op_exec_context.operator_name != MODEL_CONST_OP_NAME:
                module_id_vs_known_op_addrs_map[dynamic_graph_node.calling_module_id].add(
                    dynamic_graph_node.op_exec_context.op_address
                )

        module_id_vs_sorted_scopes_map = {
            k: list(sorted([s.scope_in_model for s in v], key=str)) for k, v in module_id_vs_known_op_addrs_map.items()
        }

        nncf_graph = PTNNCFGraph()
        for dynamic_graph_node in dynamic_graph.get_all_nodes():
            op_address = dynamic_graph_node.op_exec_context.op_address

            metatype = PT_OPERATOR_METATYPES.get_operator_metatype_by_op_name(op_address.operator_name)
            if metatype.get_subtypes():
                subtype = metatype.determine_subtype(
                    dynamic_graph_node.layer_attributes, functions_kwargs=dynamic_graph_node.__dict__
                )
            else:
                subtype = None
            if subtype is not None:
                metatype = subtype

            is_integer_input = False
            if metatype in INPUT_NOOP_METATYPES:
                is_integer_input = dynamic_graph.is_integer_input_node(dynamic_graph_node)

            is_shared = False
            if not traced_parameters:
                is_shared = len(module_id_vs_sorted_scopes_map[dynamic_graph_node.calling_module_id]) > 1
            canonical_scope = module_id_vs_sorted_scopes_map[dynamic_graph_node.calling_module_id][0]

            node_name = str(op_address)
            if metatype in CONST_NOOP_METATYPES:
                node_name = dynamic_graph_node.layer_attributes.name

            nncf_graph.add_nncf_node(
                node_name=node_name,
                node_type=op_address.operator_name,
                node_metatype=metatype,
                layer_attributes=dynamic_graph_node.layer_attributes,
                node_id_override=dynamic_graph_node.node_id,
                layer_name=str(canonical_scope),
                ignored_algorithms=dynamic_graph_node.ignored_algorithms,
                is_in_iteration_scope=dynamic_graph_node.is_in_iteration_scope,
                is_integer_input=is_integer_input,
                is_shared=is_shared,
            )

        for dynamic_graph_edge in dynamic_graph.get_all_edges():
            nncf_graph.add_edge_between_nncf_nodes(
                from_node_id=dynamic_graph_edge.from_node_id,
                to_node_id=dynamic_graph_edge.to_node_id,
                tensor_shape=dynamic_graph_edge.activation_shape,
                input_port_id=dynamic_graph_edge.input_port_id,
                output_port_id=dynamic_graph_edge.output_port_id,
                dtype=dynamic_graph_edge.dtype,
                parallel_input_port_ids=dynamic_graph_edge.parallel_input_port_ids,
            )

        set_nodes_attributes_in_nncf_graph(nncf_graph)
        if traced_parameters:
            propagate_is_shared_attribute_from_constant_nodes(nncf_graph)
        return nncf_graph


def _propagate_true_for_is_shared_attribute(node: NNCFNode, graph: NNCFGraph, val: bool) -> None:
    """
    Propagates the is_shared attribute through specific nodes in an NNCFGraph.

    :param node: The start NNCFNode to process.
    :param graph: The NNCFGraph instance.
    :param val: Propagated value for is_shared.
    """
    node.attributes[NNCFNode.IS_SHARED_ATTR] = val
    if node.metatype in CONST_NOOP_METATYPES or node.node_type in QUANTIZE_NODE_TYPES:
        for next_node in graph.get_next_nodes(node):
            _propagate_true_for_is_shared_attribute(next_node, graph, val)


def propagate_is_shared_attribute_from_constant_nodes(graph: NNCFGraph) -> None:
    """
    Detect shared constant nodes that used in several operations and propagate is_shared node attributes nodes.
    For constant nodes the is_shared attribute set to True if constant used in multiple operations.
    For other types of node with the is_shared attribute set to True if operation which associated
    with node used shared constant.

    :param graph: NNCGraph instance.
    """
    for const_node in graph.get_nodes_by_metatypes(CONST_NOOP_METATYPES):
        next_nodes = graph.get_next_nodes(const_node)
        if len(next_nodes) > 1:
            _propagate_true_for_is_shared_attribute(const_node, graph, True)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
from typing import Callable, List, Optional, TypeVar

from nncf import Dataset
from nncf.common.graph.graph import NNCFGraph
from nncf.common.quantization.structs import QuantizationPreset
from nncf.common.tensor_statistics.statistic_point import StatisticPointsContainer
from nncf.common.utils.backend import BackendType
from nncf.parameters import ModelType
from nncf.parameters import QuantizationMode
from nncf.parameters import TargetDevice
from nncf.quantization.advanced_parameters import AdvancedQuantizationParameters
from nncf.quantization.algorithms.algorithm import Algorithm
from nncf.quantization.algorithms.post_training.pipeline import create_ptq_pipeline
from nncf.scopes import IgnoredScope

TModel = TypeVar("TModel")
TPass = Callable[[TModel], TModel]


class PostTrainingQuantization(Algorithm):
    """
    Implements Post-Training Quantization algorithm, which basically includes:
    1) ChannelAlignment
    2) MinMaxQuantization
    3) FastBiasCorrection or BiasCorrection
    """

    def __init__(
        self,
        mode: Optional[QuantizationMode] = None,
        preset: Optional[QuantizationPreset] = None,
        target_device: TargetDevice = TargetDevice.ANY,
        subset_size: int = 300,
        fast_bias_correction: bool = True,
        model_type: Optional[ModelType] = None,
        ignored_scope: Optional[IgnoredScope] = None,
        advanced_parameters: Optional[AdvancedQuantizationParameters] = None,
    ):
        """
        :param mode: Special quantization mode that specify different ways of the optimization.
        :param preset: A preset controls the quantization mode (symmetric and asymmetric).
            It can take the following values:
            - `performance`: Symmetric quantization of weights and activations.
            - `mixed`: Symmetric quantization of weights and asymmetric quantization of activations.
            Default value is None. In this case, `mixed` preset is used for `transformer`
            model type otherwise `performace`.
        :param target_device: A target device the specificity of which will be taken
            into account while compressing in order to obtain the best performance
            for this type of device.
        :param subset_size: Size of a subset to calculate activations
            statistics used for quantization.
        :param fast_bias_correction: Setting this option to `False` enables a different
            bias correction method which is more accurate, in general, and takes
            more time but requires less memory.
        :param model_type: Model type is needed to specify additional patterns
            in the model. Supported only `transformer` now.
        :param ignored_scope: An ignored scope that defined the list of model control
            flow graph nodes to be ignored during quantization.
        :param advanced_parameters: Advanced quantization parameters for
            fine-tuning the quantization algorithm
        """
        self._pipeline = create_ptq_pipeline(
            mode=mode,
            preset=preset,
            target_device=target_device,
            subset_size=subset_size,
            fast_bias_correction=fast_bias_correction,
            model_type=model_type,
            ignored_scope=ignored_scope,
            advanced_parameters=advanced_parameters,
        )

    @property
    def available_backends(self) -> List[BackendType]:
        backends = set(BackendType)
        for algorithm in itertools.chain.from_iterable(self._pipeline.pipeline_steps):
            backends = backends.intersection(algorithm.available_backends)
        return list(backends)

    def get_statistic_points(self, model: TModel, graph: NNCFGraph) -> StatisticPointsContainer:
        return self._pipeline.get_statistic_points_for_step(0, model, graph)

    def apply(
        self,
        model: TModel,
        graph: NNCFGraph,
        statistic_points: Optional[StatisticPointsContainer] = None,
        dataset: Optional[Dataset] = None,
    ) -> TModel:
        if dataset is None and len(self._pipeline.pipeline_steps) > 1:
            raise ValueError(
                "A dataset is required for the post-training quantization "
                "algorithm to collect statistics for intermediate models."
            )

        step_index_to_statistics = None
        if statistic_points:
            step_index_to_statistics = {0: statistic_points}

        return self._pipeline.run_from_step(model, dataset, graph, 0, step_index_to_statistics)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from nncf.common.graph.patterns import GraphPattern
from nncf.common.graph.patterns import merge_two_types_of_operations

LINEAR_OPERATIONS = {
    GraphPattern.METATYPE_ATTR: [
        "linear",
        "conv1d",
        "conv2d",
        "conv3d",
        "conv_transpose1d",
        "conv_transpose2d",
        "conv_transpose3d",
        "deform_conv2d",
        "addmm",
        "bmm",
        "matmul",
        "mm",
        "baddbmm",
    ],
    GraphPattern.LABEL_ATTR: "LINEAR",
}

BATCH_NORMALIZATION_OPERATIONS = {
    GraphPattern.METATYPE_ATTR: ["batch_norm", "batch_norm1d", "batch_norm2d", "batch_norm3d"],
    GraphPattern.LABEL_ATTR: "BATCH_NORMALIZATION",
}

GROUP_NORMALIZATION_OPERATIONS = {
    GraphPattern.METATYPE_ATTR: ["group_norm"],
    GraphPattern.LABEL_ATTR: "GROUP_NORMALIZATION",
}

LAYER_NORMALIZATION_OPERATIONS = {
    GraphPattern.METATYPE_ATTR: ["layer_norm"],
    GraphPattern.LABEL_ATTR: "LAYER_NORMALIZATION",
}

RELU_OPERATIONS = {GraphPattern.METATYPE_ATTR: ["relu", "relu_", "hardtanh"], GraphPattern.LABEL_ATTR: "RELU"}

NON_RELU_ACTIVATIONS_OPERATIONS = {
    GraphPattern.METATYPE_ATTR: [
        "elu",
        "elu_",
        "prelu",
        "leaky_relu",
        "sigmoid",
        "gelu",
        "silu",
        "hardsigmoid",
        "hardswish",
    ],
    GraphPattern.LABEL_ATTR: "NON_RELU_ACTIVATIONS",
}

ATOMIC_ACTIVATIONS_OPERATIONS = merge_two_types_of_operations(
    RELU_OPERATIONS, NON_RELU_ACTIVATIONS_OPERATIONS, "ATOMIC_ACTIVATIONS"
)

ARITHMETIC_OPERATIONS = {
    GraphPattern.METATYPE_ATTR: ["__iadd__", "__add__", "__mul__", "__rmul__", "__truediv__"],
    GraphPattern.LABEL_ATTR: "ARITHMETIC",
}

# This type may be useful in the future

POOLING_OPERATIONS = {
    GraphPattern.METATYPE_ATTR: ["adaptive_avg_pool2d", "adaptive_avg_pool3d", "avg_pool2d", "avg_pool3d"],
    GraphPattern.LABEL_ATTR: "POOLING",
}



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging

from nncf.api.compression import CompressionStage
from nncf.common.schedulers import BaseCompressionScheduler
from nncf.common.utils.registry import Registry
from nncf.config.schemata.defaults import ACTIVATIONS_QUANT_START_EPOCH
from nncf.config.schemata.defaults import WEIGHTS_QUANT_START_EPOCH

logger = logging.getLogger(__name__)

QUANTIZATION_SCHEDULERS = Registry("quantization_schedulers")


@QUANTIZATION_SCHEDULERS.register("staged")
class StagedQuantizationScheduler(BaseCompressionScheduler):
    def __init__(self, quantization_ctrl: "QuantizationController", params=None):  # noqa: F821
        super().__init__()
        if params is None:
            params = {}
        self.algo = quantization_ctrl
        self.activations_quant_start_epoch = params.get("activations_quant_start_epoch", ACTIVATIONS_QUANT_START_EPOCH)
        self.weights_quant_start_epoch = params.get("weights_quant_start_epoch", WEIGHTS_QUANT_START_EPOCH)
        self._set_quantization_status()

    def epoch_step(self, next_epoch=None):
        super().epoch_step(next_epoch)
        should_call_init = False
        if self.current_epoch == self.activations_quant_start_epoch:
            logger.info("Enabled quantization of activations")
            self.algo.enable_activation_quantization()
            should_call_init = True

        if self.current_epoch == self.weights_quant_start_epoch:
            logger.info("Enabled quantization of weights")
            self.algo.enable_weight_quantization()
            should_call_init = True

        if should_call_init:
            self.algo.init_range()

    def load_state(self, state):
        super().load_state(state)
        # Just enables/disables quantizers without calling initialization of ranges, because it's called on epoch_step
        # in the end of previous epoch before saving the scheduler's state dict.
        self._set_quantization_status()

    def _set_quantization_status(self):
        if max(self.current_epoch, 0) >= self.activations_quant_start_epoch:
            self.algo.enable_activation_quantization()
            logger.info("Enabled quantization of activations")
        else:
            self.algo.disable_activation_quantization()
            logger.info("Disabled quantization of activations")
        if max(self.current_epoch, 0) >= self.weights_quant_start_epoch:
            self.algo.enable_weight_quantization()
            logger.info("Enabled quantization of weights")
        else:
            self.algo.disable_weight_quantization()
            logger.info("Disabled quantization of weights")

    def _calc_density_level(self):
        raise NotImplementedError

    def compression_stage(self):
        is_activations_enabled = self.current_epoch >= self.activations_quant_start_epoch
        is_weights_enabled = self.current_epoch >= self.weights_quant_start_epoch
        if is_activations_enabled and is_weights_enabled:
            return CompressionStage.FULLY_COMPRESSED
        if not is_activations_enabled and not is_weights_enabled:
            return CompressionStage.UNCOMPRESSED
        return CompressionStage.PARTIALLY_COMPRESSED



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Type

from nncf.torch.quantization.precision_init.autoq_init import AutoQPrecisionInitializer
from nncf.torch.quantization.precision_init.base_init import BasePrecisionInitializer
from nncf.torch.quantization.precision_init.hawq_init import HAWQPrecisionInitializer
from nncf.torch.quantization.precision_init.manual_init import ManualPrecisionInitializer


class PrecisionInitializerFactory:
    @staticmethod
    def create(init_type: str) -> Type[BasePrecisionInitializer]:
        if init_type == "manual":
            return ManualPrecisionInitializer
        if init_type == "hawq":
            return HAWQPrecisionInitializer
        if init_type == "autoq":
            return AutoQPrecisionInitializer
        raise NotImplementedError



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from abc import ABC
from abc import abstractmethod
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, TypeVar

import nncf
from nncf import NNCFConfig
from nncf.api.compression import CompressionAlgorithmBuilder
from nncf.api.compression import CompressionAlgorithmController
from nncf.common.logging import nncf_logger
from nncf.common.schedulers import StubCompressionScheduler
from nncf.common.utils.api_marker import api
from nncf.common.utils.backend import BackendType
from nncf.common.utils.backend import get_backend
from nncf.common.utils.os import fail_if_symlink
from nncf.common.utils.registry import Registry
from nncf.config.extractors import BNAdaptDataLoaderNotFoundError
from nncf.config.extractors import extract_algo_specific_config
from nncf.config.extractors import extract_bn_adaptation_init_params
from nncf.config.extractors import has_bn_section
from nncf.config.schemata.defaults import VALIDATE_SCOPES

TModel = TypeVar("TModel")

NO_COMPRESSION_ALGORITHM_NAME = "NoCompressionAlgorithm"


class BaseControllerStateNames:
    LOSS = "loss_state"
    SCHEDULER = "scheduler_state"
    COMPRESSION_STAGE = "compression_stage"
    COMPRESSION_LEVEL = "compression_level"


@api()
class BaseCompressionAlgorithmController(CompressionAlgorithmController, ABC):
    """
    Contains the implementation of the basic functionality of the compression controller.
    """

    BUILDER_STATE = "builder_state"
    CONTROLLER_STATE = "ctrl_state"
    _state_names = BaseControllerStateNames

    def __init__(self, target_model: TModel):
        """
        Initializes the internal state of the compression algorithm controller.

        :param target_model: The model with additional modifications necessary
            to enable algorithm-specific compression during fine-tuning built
            by the `CompressionAlgorithmBuilder`.
        """
        super().__init__(target_model)
        self._name = None
        self._builder_state = None

    @property
    def name(self):
        if self._name is None:
            raise nncf.InternalError("Internal error: name of the controller is not set!")
        return self._name

    @property
    def compression_rate(self) -> float:
        return None

    @compression_rate.setter
    def compression_rate(self) -> float:
        pass

    def export_model(
        self,
        save_path: str,
        save_format: Optional[str] = None,
        input_names: Optional[List[str]] = None,
        output_names: Optional[List[str]] = None,
        model_args: Optional[Tuple[Any, ...]] = None,
    ) -> None:
        """
        Exports the compressed model to the specified format for deployment.

        Makes method-specific preparations of the model, (e.g. removing auxiliary
        layers that were used for the model compression), then exports the model to
        the specified path.

        :param save_path: The path where the model will be saved.
        :param save_format: Saving format. The default format will
            be used if `save_format` is not specified.
        :param input_names: Names to be assigned to the input tensors of the model.
        :param output_names: Names to be assigned to the output tensors of the model.
        :param model_args: Tuple of additional positional and keyword arguments
            which are required for the model's forward during export. Should be
            specified in the following format:
                - (a, b, {'x': None, 'y': y}) for positional and keyword arguments.
                - (a, b, {}) for positional arguments only.
                - ({'x': None, 'y': y},) for keyword arguments only.
        """
        fail_if_symlink(Path(save_path))
        self.prepare_for_export()
        backend = get_backend(self.model)
        if backend is BackendType.TENSORFLOW:
            from nncf.tensorflow.exporter import TFExporter

            exporter = TFExporter(self.model, input_names, output_names, model_args)
        else:
            assert backend is BackendType.TORCH
            from nncf.torch.exporter import PTExporter

            exporter = PTExporter(self.model, input_names, output_names, model_args)
        if save_format is not None:
            exporter.export_model(save_path, save_format)
        else:
            exporter.export_model(save_path)

    def disable_scheduler(self) -> None:
        self._scheduler = StubCompressionScheduler()

    def set_builder_state_with_name(self, name: str, builder_state: Dict):
        """
        Sets state of the builder and the corresponding algorithm name. Should be called by the builder to set its
        state and registered algorithm key.

        :param name: algorithm name, the string that was used to register the builder
        :param builder_state: state of the builder
        """
        self._name = name
        self._builder_state = builder_state

    def load_state(self, state: Dict[str, Dict[str, Any]]) -> None:
        """
        Loads the compression controller state from the map of algorithm name to the dictionary with state attributes.

        :param state: map of the algorithm name to the dictionary with the corresponding state attributes.
        """
        if self.name in state:
            algo_state = state[self.name]
            if self._state_names.COMPRESSION_STAGE in state:
                compression_stage = state[self._state_names.COMPRESSION_STAGE]
                if self.compression_stage() != compression_stage:
                    nncf_logger.warning(
                        f"Current CompressionStage ({self.compression_stage()}) of the compression controller "
                        f"does not correspond to the value found in the checkpoint ({compression_stage})"
                    )
            self.loss.load_state(algo_state[self._state_names.LOSS])
            self.scheduler.load_state(algo_state[self._state_names.SCHEDULER])

    def get_state(self) -> Dict[str, Dict[str, Any]]:
        """
        Returns compression controller state, which is the map of the algorithm name to the dictionary with the
        corresponding state attributes.

        :return: The compression controller state.
        """
        return {
            self.name: {
                self._state_names.LOSS: self.loss.get_state(),
                self._state_names.SCHEDULER: self.scheduler.get_state(),
                self._state_names.COMPRESSION_STAGE: self.compression_stage(),
            }
        }

    def get_compression_state(self) -> Dict[str, Any]:
        """
        Returns compression state - builder and controller state.
        This state should be used to resume compression via `compression_state` argument of `create_compressed_model`
        method.

        :return: The compression state.
        """
        if self._builder_state is None:
            raise nncf.InternalError("Internal error: builder state is not set for the controller")

        return {self.BUILDER_STATE: self._builder_state, self.CONTROLLER_STATE: self.get_state()}

    @property
    def maximal_compression_rate(self) -> float:
        return 1.0


class BaseCompressionAlgorithmBuilder(CompressionAlgorithmBuilder):
    """
    Contains the implementation of the basic functionality of a single compression algorithm builder.
    """

    def __init__(self, config: NNCFConfig, should_init: bool = True):
        """
        Initializes internal state of the compression algorithm builder

        :param config: The top-level NNCFConfig object (i.e. parsed from a .json and extended with
            all necessary objects required for compression such as initialization data loaders).
        :param should_init: If False, trainable parameter initialization will be
            skipped during building.
        """
        super().__init__()
        self.config = config
        self.should_init = should_init
        self._algo_config = self._get_algo_specific_config_section()

        self.validate_scopes = self._algo_config.get("validate_scopes", VALIDATE_SCOPES)

        self.ignored_scopes = self.config.get("ignored_scopes")
        if "ignored_scopes" in self._algo_config:
            algo_ignored_scopes = self._algo_config["ignored_scopes"]
            if self.ignored_scopes is not None:
                self.ignored_scopes.extend(algo_ignored_scopes)
            else:
                self.ignored_scopes = algo_ignored_scopes

        self._global_target_scopes = self.config.get("target_scopes")
        self.target_scopes = self._global_target_scopes
        if "target_scopes" in self._algo_config:
            algo_target_scopes = self._algo_config["target_scopes"]
            if self.target_scopes is None:
                self.target_scopes = algo_target_scopes

    def _get_algo_specific_config_section(self) -> Dict:
        return extract_algo_specific_config(self.config, self.name)

    @property
    def name(self) -> str:
        return getattr(self, Registry.REGISTERED_NAME_ATTR, "NOT_REGISTERED_" + self.__class__.__name__)

    def load_state(self, state: Dict[str, Any]) -> None:
        """
        Initializes object from the state.

        :param state: Output of `get_state()` method.
        """
        if self.name in state:
            self._load_state_without_name(state[self.name])

    def get_state(self) -> Dict[str, Any]:
        """
        Returns a dictionary with Python data structures (dict, list, tuple, str, int, float, True, False, None) that
        represents state of the object.

        :return: state of the object
        """
        return {self.name: self._get_state_without_name()}

    @abstractmethod
    def _build_controller(self, model: TModel) -> BaseCompressionAlgorithmController:
        """
        Simple implementation of building controller without setting builder state and loading controller's one.

        :param model: The model with additional modifications necessary to enable
            algorithm-specific compression during fine-tuning.
        :return: The instance of the `BaseCompressionAlgorithmController`.
        """

    def build_controller(self, model: TModel) -> BaseCompressionAlgorithmController:
        """
        Builds `BaseCompressionAlgorithmController` to handle the additional modules,
        parameters, and hooks inserted into the model to enable algorithm-specific
        compression.

        :param model: The model with additional modifications necessary to enable
            algorithm-specific compression during fine-tuning.
        :return: The instance of the `BaseCompressionAlgorithmController`.
        """
        ctrl = self._build_controller(model)
        ctrl.set_builder_state_with_name(self.name, self.get_state())
        return ctrl

    @abstractmethod
    def _load_state_without_name(self, state_without_name: Dict[str, Any]):
        """
        Implementation of load state that takes state without builder name.

        :param state_without_name: Output of `_get_state_without_name()` method.
        """

    @abstractmethod
    def _get_state_without_name(self) -> Dict[str, Any]:
        """
        Implementation of get_state that returns state without builder name.

        :return: Returns a dictionary with Python data structures
            (dict, list, tuple, str, int, float, True, False, None) that represents state of the object.
        """

    def _parse_bn_adapt_params(self) -> Optional[Dict]:
        try:
            return extract_bn_adaptation_init_params(self.config, self.name)
        except BNAdaptDataLoaderNotFoundError as e:
            if not has_bn_section(self.config, self.name):
                nncf_logger.info(
                    "Data loader for batchnorm adaptation not found in NNCFConfig and no explicit batchnorm adaptation"
                    "parameters were passed in config - will not perform batchnorm adaptation.\n"
                    "It is recommended to do batchnorm adaptation after creating a compressed model - use "
                    "`register_default_init_args` or `nncf.NNCFConfig.register_extra_structs` directly to register a "
                    "dataloader and NNCF will do batchnorm adaptation automatically at compressed model creation."
                )
                return None
            raise e



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re
from copy import deepcopy
from typing import List

import nncf


class ScopeElement:
    def __init__(self, calling_module_class_name: str, calling_field_name: str = None):
        self.calling_module_class_name = calling_module_class_name
        self.calling_field_name = calling_field_name

    def __str__(self):
        if self.calling_field_name is None:
            return self.calling_module_class_name
        return "{cls}[{name}]".format(cls=self.calling_module_class_name, name=self.calling_field_name)

    def __eq__(self, other: "ScopeElement"):
        return (self.calling_module_class_name == other.calling_module_class_name) and (
            self.calling_field_name == other.calling_field_name
        )

    def __hash__(self):
        return hash((self.calling_module_class_name, self.calling_field_name))

    @staticmethod
    def from_str(string: str):
        matches = re.search(r"(.*)\[(.*)\]|(.*)", string)
        if matches is None:
            raise nncf.InternalError("Invalid scope element string")
        if matches.groups()[0] is None and matches.groups()[1] is None:
            return ScopeElement(matches.groups()[2])
        if matches.groups()[0] is not None and matches.groups()[1] is not None:
            return ScopeElement(matches.groups()[0], matches.groups()[1])
        raise nncf.InternalError("Could not parse the scope element string")


class Scope:
    def __init__(self, scope_elements: List[ScopeElement] = None):
        if scope_elements is not None:
            self.scope_elements = scope_elements
        else:
            self.scope_elements = []

    def __str__(self):
        return "/".join([str(scope_el) for scope_el in self.scope_elements])

    def __repr__(self):
        return str(self)

    def __hash__(self):
        return hash(str(self))

    def __eq__(self, other: "Scope"):
        return self.scope_elements == other.scope_elements

    def __getitem__(self, key) -> ScopeElement:
        return self.scope_elements[key]

    def __contains__(self, item: "Scope"):
        """Idiom: ('A/B/C' in 'A/B') == True"""
        if len(self.scope_elements) > len(item.scope_elements):
            return False
        for i, element in enumerate(self.scope_elements):
            if element != item.scope_elements[i]:
                return False
        return True

    def __add__(self, rhs: "Scope") -> "Scope":
        init_list = self.scope_elements + rhs.scope_elements
        return Scope(init_list)

    def copy(self) -> "Scope":
        return Scope(deepcopy(self.scope_elements))

    def push(self, scope_element: ScopeElement):
        self.scope_elements.append(scope_element)

    def pop(self) -> ScopeElement:
        return self.scope_elements.pop()

    @staticmethod
    def from_str(string: str) -> "Scope":
        if string:
            elts = string.split("/")
        else:
            elts = []
        return Scope([ScopeElement.from_str(s) for s in elts])

    def get_iteration_scopes(self) -> List[str]:
        results = []
        from nncf.torch.layers import ITERATION_MODULES

        for scope_element in self.scope_elements:
            if scope_element.calling_module_class_name in ITERATION_MODULES.registry_dict:
                results.append(scope_element.calling_module_class_name)
        return results



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
from copy import deepcopy
from itertools import islice
from typing import List

import numpy as np
import tensorflow as tf

from nncf.common.logging.progress_bar import ProgressBar
from nncf.common.quantization.initialization.range import RangeInitCollectorParams
from nncf.common.quantization.initialization.range import RangeInitConfig
from nncf.common.quantization.initialization.range import RangeInitParams
from nncf.common.quantization.structs import QuantizerGroup
from nncf.common.scopes import should_consider_scope
from nncf.common.tensor_statistics.collectors import ReductionAxes
from nncf.common.tensor_statistics.collectors import TensorStatisticCollectorBase
from nncf.config.schemata.defaults import MAX_PERCENTILE
from nncf.config.schemata.defaults import MIN_PERCENTILE
from nncf.tensorflow.layers.custom_objects import NNCF_QUANTIZATION_OPERATIONS
from nncf.tensorflow.layers.data_layout import get_channel_axis
from nncf.tensorflow.layers.operation import InputType
from nncf.tensorflow.layers.wrapper import NNCFWrapper
from nncf.tensorflow.quantization.layers import FakeQuantize
from nncf.tensorflow.tensor_statistics.collectors import TFMeanMinMaxStatisticCollector
from nncf.tensorflow.tensor_statistics.collectors import TFMeanPercentileStatisticCollector
from nncf.tensorflow.tensor_statistics.collectors import TFMedianMADStatisticCollector
from nncf.tensorflow.tensor_statistics.collectors import TFMinMaxStatisticCollector
from nncf.tensorflow.tensor_statistics.collectors import TFMixedMinMaxStatisticCollector
from nncf.tensorflow.tensor_statistics.collectors import TFPercentileStatisticCollector
from nncf.tensorflow.tensor_statistics.reduction import get_reduction_shape_activations
from nncf.tensorflow.tensor_statistics.reduction import get_reduction_shape_weights
from nncf.tensorflow.tensor_statistics.statistics import tf_convert_stat_to_min_max_tensor_stat


class TFRangeInitParams(RangeInitParams):
    def get_max_num_init_steps(self) -> int:
        steps = []
        if self.global_init_config is not None:
            steps.append(self.global_init_config.num_init_samples)
        for pl_config in self.per_layer_range_init_configs:
            steps.append(pl_config.num_init_samples)
        batch_size = self.init_range_data_loader.batch_size
        return math.ceil(max(steps) / batch_size)

    def get_init_config_for_quantization_point(self, layer: tf.keras.layers.Layer, input_type: str) -> RangeInitConfig:
        if input_type == InputType.WEIGHTS:
            node_name = layer.name
            group = QuantizerGroup.WEIGHTS
        else:
            node_name = layer.name.replace("/fake_quantize", "")
            group = QuantizerGroup.ACTIVATIONS
        return self.get_init_config_for_scope_and_group(node_name, group)

    def get_init_config_for_scope_and_group(self, node_name: str, group: QuantizerGroup) -> RangeInitConfig:
        matches: List[RangeInitConfig] = []
        for pl_config in self.per_layer_range_init_configs:
            should_be_considered = should_consider_scope(
                node_name, ignored_scopes=pl_config.ignored_scopes, target_scopes=pl_config.target_scopes
            )
            if should_be_considered and (group == pl_config.target_group or pl_config.target_group is None):
                matches.append(
                    RangeInitConfig(
                        pl_config.init_type, pl_config.num_init_samples, pl_config.init_type_specific_params
                    )
                )
        if len(matches) > 1:
            raise ValueError(
                "Location {} matches more than one per-layer initialization parameter "
                "definition!".format(str(node_name))
            )
        if len(matches) == 1:
            return matches[0]
        if not matches and self.global_init_config is not None:
            return deepcopy(self.global_init_config)

        raise ValueError(
            "Location {} does not match any per-layer initialization parameter definition!".format(str(node_name))
        )


class RangeInitializer:
    def __init__(self, range_init_params: TFRangeInitParams):
        self.range_init_params = range_init_params
        self.dataset = range_init_params.init_range_data_loader
        self.num_steps = range_init_params.get_max_num_init_steps()

        self.nncf_quantization_operation_classes = NNCF_QUANTIZATION_OPERATIONS.registry_dict.values()

    @staticmethod
    def generate_stat_collector(
        reduction_shape: ReductionAxes,
        collector_params: RangeInitCollectorParams,
        init_config: RangeInitConfig,
        num_samples_to_collect_override: int = None,
    ) -> TensorStatisticCollectorBase:
        range_type = init_config.init_type
        num_samples = init_config.num_init_samples
        if num_samples_to_collect_override is not None:
            num_samples = num_samples_to_collect_override

        if range_type == "min_max":
            return TFMinMaxStatisticCollector(collector_params.use_abs_max, reduction_shape, num_samples)
        if range_type == "mixed_min_max":
            return TFMixedMinMaxStatisticCollector(
                collector_params.use_per_sample_stats(per_sample_stats=True),
                collector_params.use_abs_max,
                collector_params.use_means_of_mins,
                collector_params.use_means_of_maxs,
                reduction_shape,
                num_samples,
            )
        if range_type == "mean_min_max":
            return TFMeanMinMaxStatisticCollector(
                collector_params.use_per_sample_stats(per_sample_stats=True),
                collector_params.use_abs_max,
                reduction_shape,
                num_samples,
            )
        if range_type == "threesigma":
            return TFMedianMADStatisticCollector(reduction_shape, num_samples)
        if range_type == "percentile":
            min_percentile = init_config.init_type_specific_params.get("min_percentile", MIN_PERCENTILE)
            max_percentile = init_config.init_type_specific_params.get("max_percentile", MAX_PERCENTILE)
            return TFPercentileStatisticCollector([min_percentile, max_percentile], reduction_shape, num_samples)
        if range_type == "mean_percentile":
            min_percentile = init_config.init_type_specific_params.get("min_percentile", MIN_PERCENTILE)
            max_percentile = init_config.init_type_specific_params.get("max_percentile", MAX_PERCENTILE)
            return TFMeanPercentileStatisticCollector([min_percentile, max_percentile], reduction_shape, num_samples)
        raise ValueError(f"Range type {range_type} is not supported.")

    def _register_layer_statistics(self, layer: tf.keras.layers.Layer, layer_statistics: list, handles: list):
        channel_axes = get_channel_axis(InputType.INPUTS, "", layer)
        init_config = self.range_init_params.get_init_config_for_quantization_point(layer, InputType.INPUTS)

        is_weights = False
        collector_params = RangeInitCollectorParams(is_weights, layer.mode, layer.per_channel)
        per_sample_stats = init_config.init_type in ["mixed_min_max", "mean_min_max"]

        reduction_shape = get_reduction_shape_activations(
            layer, channel_axes, collector_params.use_per_sample_stats(per_sample_stats)
        )

        num_batches = int(np.ceil(init_config.num_init_samples / self.dataset.batch_size))

        collector = RangeInitializer.generate_stat_collector(
            reduction_shape, collector_params, init_config, num_batches
        )
        handles.append(layer.register_hook_pre_quantizer(collector.register_input))
        layer.enabled = False
        layer_statistics.append((layer, collector))

    def _register_op_statistics(self, layer: tf.keras.layers.Layer, op_statistics: list, handles: list):
        for weight_attr, ops in layer.weights_attr_ops.items():
            for op_name, op in ops.items():
                if op.__class__ in self.nncf_quantization_operation_classes:
                    channel_axes = get_channel_axis(InputType.WEIGHTS, weight_attr, layer)
                    init_config = self.range_init_params.get_init_config_for_quantization_point(
                        layer, InputType.WEIGHTS
                    )

                    is_weights = True
                    collector_params = RangeInitCollectorParams(is_weights, op.mode, op.per_channel)

                    reduction_shape = get_reduction_shape_weights(layer, weight_attr, channel_axes, op.per_channel)

                    # No need to store extra statistics in memory since weights won't change during range init
                    num_batches = 1

                    collector = RangeInitializer.generate_stat_collector(
                        reduction_shape, collector_params, init_config, num_batches
                    )
                    handles.append(op.register_hook_pre_call(collector.register_input))
                    op.enabled = False
                    op_statistics.append((layer, op_name, op, collector))

    def run(self, model: tf.keras.Model) -> None:
        layer_statistics = []
        op_statistics = []
        handles = []
        for layer in model.layers:
            if isinstance(layer, FakeQuantize):
                self._register_layer_statistics(layer, layer_statistics, handles)
            elif isinstance(layer, NNCFWrapper):
                self._register_op_statistics(layer, op_statistics, handles)

        for x, _ in ProgressBar(
            islice(self.dataset, self.num_steps), total=self.num_steps, desc="Collecting tensor statistics/data"
        ):
            model(x, training=False)

        for layer, collector in layer_statistics:
            target_stat = collector.get_statistics()
            minmax_stats = tf_convert_stat_to_min_max_tensor_stat(target_stat)
            layer.apply_range_initialization(tf.squeeze(minmax_stats.min_values), tf.squeeze(minmax_stats.max_values))
            layer.enabled = True

        for layer, op_name, op, collector in op_statistics:
            weights = layer.get_operation_weights(op_name)
            target_stat = collector.get_statistics()
            minmax_stats = tf_convert_stat_to_min_max_tensor_stat(target_stat)
            min_values = minmax_stats.min_values
            if len(min_values.shape) != 1:
                min_values = tf.squeeze(min_values)
            max_values = minmax_stats.max_values
            if len(max_values.shape) != 1:
                max_values = tf.squeeze(max_values)
            op.apply_range_initialization(weights, min_values, max_values)
            op.enabled = True

        for handle in handles:
            handle.remove()

        for x, _ in self.dataset:
            model(x, training=False)
            break



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Tuple

import nncf
from nncf.common.graph import NNCFGraph
from nncf.common.graph import NNCFNode
from nncf.common.graph import NNCFNodeName
from nncf.common.graph.layer_attributes import MultipleInputLayerAttributes
from nncf.torch.dynamic_graph.scope import Scope
from nncf.torch.graph.transformations.commands import PTTargetPoint


class PTNNCFGraph(NNCFGraph):
    def get_output_shapes_for_node(self, node_name: NNCFNodeName) -> List[Tuple]:
        node = self.get_node_by_name(node_name)
        node_key = self.get_node_key_by_id(node.node_id)
        succs = list(self._nx_graph.successors(node_key))
        edge_list = [self._nx_graph.edges[node_key, to_node_key] for to_node_key in succs]
        return [edge[NNCFGraph.ACTIVATION_SHAPE_EDGE_ATTR] for edge in edge_list]

    def get_input_shapes_for_node(self, node_name: NNCFNodeName) -> Dict[int, Tuple]:
        node = self.get_node_by_name(node_name)
        node_key = self.get_node_key_by_id(node.node_id)
        in_edges = list(self._nx_graph.in_edges(node_key))
        retval = {}
        for in_edge in in_edges:
            edge_attr_dict = self._nx_graph.edges[in_edge]
            port_id = edge_attr_dict[NNCFGraph.INPUT_PORT_ID_EDGE_ATTR]
            assert port_id not in retval
            for p in [
                port_id,
            ] + edge_attr_dict[NNCFGraph.PARALLEL_INPUT_PORT_IDS_ATTR]:
                retval[p] = edge_attr_dict[NNCFGraph.ACTIVATION_SHAPE_EDGE_ATTR]
        return retval

    def get_input_shape_for_insertion_point(self, insertion_point: PTTargetPoint) -> Tuple[int]:
        target_node_name = insertion_point.target_node_name
        if insertion_point.input_port_id is not None:
            quantizer_input_shape = self.get_input_shapes_for_node(target_node_name)[insertion_point.input_port_id]
        else:
            # Tailored for post-hook quantization and first output quantization only
            quantizer_input_shape = self.get_output_shapes_for_node(target_node_name)[0]
        return quantizer_input_shape

    def get_op_nodes_in_scope(self, scope: Scope) -> List[NNCFNode]:
        """
        Returns all NNCFNodes inside the given scope.

        :param scope: Given scope.
        :return: All NNCFNodes inside the given scope.
        """
        matching_graph_op_nodes = []
        for scope_str, nodes_in_module in self._layer_name_vs_shared_nodes.items():
            module_scope = Scope.from_str(scope_str)
            if module_scope in scope:
                matching_graph_op_nodes.extend(nodes_in_module)
        return matching_graph_op_nodes

    def get_op_nodes_with_scope(self, scope: Scope) -> List[NNCFNode]:
        """
        Returns all NNCFNodes which share the given scope.

        :param scope: Given scope.
        :return: All NNCFNodes which share the given scope.
        """
        return self._layer_name_vs_shared_nodes[str(scope)]

    def get_scope_by_node_name(self, node_name: NNCFNodeName) -> Scope:
        """
        Returns a scope which corresponds to the given NNCF node name.

        :param node_name: Given node name.
        :return: A scope which corresponds to the given NNCF node name.
        """
        matches = []
        for node_id, scope_str in self._node_ids_vs_layer_names.items():
            node = self.get_node_by_id(node_id)
            if node.node_name == node_name:
                matches.append(Scope.from_str(scope_str))
        assert len(matches) <= 1
        if not matches:
            raise nncf.InternalError("Node name {} not found in the node-vs-scope dict!".format(node_name))
        return matches[0]

    def get_nodes_with_missed_input_edges(self) -> List[NNCFNode]:
        """
        Returns a list of NNCFNodes that have at least one expected input edge missed.
        Requires MultipleInputLayerAttributes for nodes with several inputs and
        right `num_expected_input_edges` parameter setted for nncf nodes metatypes.

        :return: List of NNCFNodes that are identified as diconected.
        """
        input_nodes = set()
        for node in self.get_all_nodes():
            num_expected_input_edges = None
            if hasattr(node.metatype, "num_expected_input_edges"):
                num_expected_input_edges = node.metatype.num_expected_input_edges
            if node.layer_attributes is not None and isinstance(node.layer_attributes, MultipleInputLayerAttributes):
                num_expected_input_edges = node.layer_attributes.num_inputs
            if num_expected_input_edges:
                input_edges = self.get_input_edges(node)
                if len(input_edges) < num_expected_input_edges:
                    # If node has missed input edges we assume this node is an input node
                    # that was disconected from an activation input.
                    input_nodes.add(node)
        return list(input_nodes)



# Copyright (c) 2024 Intel Corporation
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from copy import deepcopy
from typing import List, Set

import numpy as np
import tensorflow as tf

from nncf import NNCFConfig
from nncf.common.accuracy_aware_training.training_loop import ADAPTIVE_COMPRESSION_CONTROLLERS
from nncf.common.graph.transformations.commands import TransformationPriority
from nncf.common.schedulers import StubCompressionScheduler
from nncf.common.scopes import check_scopes_in_graph
from nncf.common.scopes import should_consider_scope
from nncf.common.sparsity.schedulers import SPARSITY_SCHEDULERS
from nncf.common.sparsity.schedulers import SparsityScheduler
from nncf.common.sparsity.statistics import RBSparsityStatistics
from nncf.common.statistics import NNCFStatistics
from nncf.common.utils.api_marker import api
from nncf.config.extractors import extract_algo_specific_config
from nncf.config.schemata.defaults import SPARSITY_INIT
from nncf.config.schemata.defaults import SPARSITY_LEVEL_SETTING_MODE
from nncf.tensorflow.algorithm_selector import TF_COMPRESSION_ALGORITHMS
from nncf.tensorflow.api.compression import TFCompressionAlgorithmBuilder
from nncf.tensorflow.graph.converter import TFModelConverterFactory
from nncf.tensorflow.graph.transformations.commands import TFInsertionCommand
from nncf.tensorflow.graph.transformations.commands import TFLayerWeight
from nncf.tensorflow.graph.transformations.layout import TFTransformationLayout
from nncf.tensorflow.graph.utils import get_nncf_operations
from nncf.tensorflow.graph.utils import get_original_name_and_instance_idx
from nncf.tensorflow.sparsity.base_algorithm import SPARSITY_LAYER_METATYPES
from nncf.tensorflow.sparsity.base_algorithm import BaseSparsityController
from nncf.tensorflow.sparsity.collector import TFSparseModelStatisticsCollector
from nncf.tensorflow.sparsity.rb.loss import SparseLoss
from nncf.tensorflow.sparsity.rb.operation import RBSparsifyingWeight


@TF_COMPRESSION_ALGORITHMS.register("rb_sparsity")
class RBSparsityBuilder(TFCompressionAlgorithmBuilder):
    def __init__(self, config: NNCFConfig, should_init: bool = True):
        super().__init__(config, should_init)
        self.ignored_scopes = self._algo_config.get("ignored_scopes", [])
        self._op_names = []

    def get_transformation_layout(self, model: tf.keras.Model) -> TFTransformationLayout:
        converter = TFModelConverterFactory.create(model)
        nncf_graph = converter.convert()

        check_scopes_in_graph(nncf_graph, self.ignored_scopes, self.target_scopes, self.validate_scopes)

        transformations = TFTransformationLayout()

        processed_shared_layer_names: Set[str] = set()

        for node in nncf_graph.get_all_nodes():
            if node.is_shared():
                target_layer_name, _ = get_original_name_and_instance_idx(node.node_name)
                if target_layer_name in processed_shared_layer_names:
                    continue
                processed_shared_layer_names.add(target_layer_name)

            if not (
                node.metatype in SPARSITY_LAYER_METATYPES
                and should_consider_scope(node.node_name, ignored_scopes=self.ignored_scopes)
            ):
                continue

            _, layer_info = converter.get_layer_info_for_node(node.node_name)
            for weight_def in node.metatype.weight_definitions:
                op_name = self._get_rb_sparsity_operation_name(node.node_name, weight_def.weight_attr_name)
                self._op_names.append(op_name)

                transformations.register(
                    TFInsertionCommand(
                        target_point=TFLayerWeight(layer_info.layer_name, weight_def.weight_attr_name),
                        callable_object=RBSparsifyingWeight(op_name),
                        priority=TransformationPriority.SPARSIFICATION_PRIORITY,
                    )
                )

        return transformations

    def _get_rb_sparsity_operation_name(self, layer_name: str, weight_attr_name: str) -> str:
        return f"{layer_name}_{weight_attr_name}_rb_sparsity_weight"

    def _build_controller(self, model: tf.keras.Model) -> "RBSparsityController":
        """
        Simple implementation of building controller without setting builder state and loading controller's one.
        Should be called once the compressed model target_model is fully constructed.

        :param model: The model with additional modifications necessary to enable
            algorithm-specific compression during fine-tuning.
        :return: The instance of the `RBSparsityController`.
        """

        return RBSparsityController(model, self.config, self._op_names)

    def initialize(self, model: tf.keras.Model) -> None:
        pass


@api()
@ADAPTIVE_COMPRESSION_CONTROLLERS.register("tf_rb_sparsity")
class RBSparsityController(BaseSparsityController):
    """
    Controller class for regularization-based (RB) sparsity in TF.
    """

    def __init__(self, target_model, config: NNCFConfig, op_names: List[str]):
        super().__init__(target_model, op_names)
        algo_config = extract_algo_specific_config(config, "rb_sparsity")
        sparsity_init = algo_config.get("sparsity_init", SPARSITY_INIT)
        params = deepcopy(algo_config.get("params", {}))
        params["sparsity_init"] = sparsity_init
        sparsity_level_mode = params.get("sparsity_level_setting_mode", SPARSITY_LEVEL_SETTING_MODE)

        if sparsity_level_mode == "local":
            raise NotImplementedError("RB sparsity algorithm do not support local sparsity loss")

        target_ops = []
        for wrapped_layer, _, op in get_nncf_operations(self.model, self._op_names):
            target_ops.append((op, wrapped_layer.get_operation_weights(op.name)))

        self._loss = SparseLoss(target_ops)
        schedule_type = params.get("schedule", "exponential")

        if schedule_type == "adaptive":
            raise NotImplementedError("RB sparsity algorithm do not support adaptive scheduler")

        scheduler_cls = SPARSITY_SCHEDULERS.get(schedule_type)
        self._scheduler = scheduler_cls(self, params)
        self.set_sparsity_level(sparsity_init)

    @property
    def scheduler(self) -> SparsityScheduler:
        return self._scheduler

    @property
    def loss(self) -> SparseLoss:
        return self._loss

    def set_sparsity_level(self, sparsity_level):
        self._loss.set_target_sparsity_loss(sparsity_level)

    @property
    def current_sparsity_level(self) -> float:
        # TODO: align with torch where it currently shows the sparsity level as reported by loss object.
        #  TF does not seem to have this functionality in its SparseLoss right now.
        return self.scheduler.current_sparsity_level

    def freeze(self):
        self._loss.disable()

    def statistics(self, quickly_collected_only: bool = False) -> NNCFStatistics:
        collector = TFSparseModelStatisticsCollector(self.model, self._op_names)
        model_stats = collector.collect()

        sparse_prob_sum = 0.0
        num_weights = 0
        for wrapped_layer, _, op in get_nncf_operations(self.model, self._op_names):
            operation_weights = wrapped_layer.get_operation_weights(op.name)
            mask = op.get_mask(operation_weights)
            sparse_prob_sum += tf.math.reduce_sum(tf.math.sigmoid(mask)).numpy().item()
            num_weights += np.prod(mask.shape.as_list()).item()
        mean_sparse_prob = 1.0 - (sparse_prob_sum / num_weights)

        target_sparsity_level = self.scheduler.current_sparsity_level

        stats = RBSparsityStatistics(model_stats, target_sparsity_level, mean_sparse_prob)

        nncf_stats = NNCFStatistics()
        nncf_stats.register("rb_sparsity", stats)
        return nncf_stats

    @property
    def compression_rate(self) -> float:
        return self._loss.target_sparsity_rate

    @compression_rate.setter
    def compression_rate(self, compression_rate: float) -> None:
        self.set_sparsity_level(compression_rate)

    def disable_scheduler(self):
        self._scheduler = StubCompressionScheduler()
